# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
from itertools import repeat
import numpy as np
from astropy import units as u
from astropy.table import Table
import gammapy.utils.parallel as parallel
from gammapy.datasets import Datasets
from gammapy.datasets.actors import DatasetsActor
from gammapy.datasets.flux_points import _get_reference_model
from gammapy.maps import MapAxis
from gammapy.modeling import Fit
from ..flux import FluxEstimator
from .core import FluxPoints

log = logging.getLogger(__name__)

__all__ = ["FluxPointsEstimator"]


class FluxPointsEstimator(FluxEstimator, parallel.ParallelMixin):
    """Flux points estimator.

    Estimates flux points for a given list of datasets, energies and spectral model.

    To estimate the flux point the amplitude of the reference spectral model is
    fitted within the energy range defined by the energy group. This is done for
    each group independently. The amplitude is re-normalized using the "norm" parameter,
    which specifies the deviation of the flux from the reference model in this
    energy group. See https://gamma-astro-data-formats.readthedocs.io/en/latest/spectra/binned_likelihoods/index.html
    for details.

    The method is also described in the `Fermi-LAT catalog paper <https://ui.adsabs.harvard.edu/abs/2015ApJS..218...23A>`__
    or the `H.E.S.S. Galactic Plane Survey paper <https://ui.adsabs.harvard.edu/abs/2018A%26A...612A...1H>`__

    Parameters
    ----------
    source : str or int
        For which source in the model to compute the flux points.
    n_sigma : float, optional
        Number of sigma to use for asymmetric error computation. Must be a positive value.
        Default is 1.
    n_sigma_ul : float, optional
        Number of sigma to use for upper limit computation. Must be a positive value.
        Default is 2.
    n_sigma_sensitivity : float, optional
        Sigma to use for sensitivity computation. Must be a positive value.
        Default is 5.
    selection_optional : list of str, optional
        Which additional quantities to estimate. Available options are:

            * "all": all the optional steps are executed.
            * "errn-errp": estimate asymmetric errors on flux.
            * "ul": estimate upper limits.
            * "scan": estimate fit statistic profiles.
            * "sensitivity": estimate sensitivity for a given significance.

        Default is None so the optional steps are not executed.
    energy_edges : list of `~astropy.units.Quantity`, optional
        Edges of the flux points energy bins. The resulting bin edges won't be exactly equal to the input ones,
        but rather the closest values to the energy axis edges of the parent dataset.
        Default is [1, 10] TeV.
    fit : `~gammapy.modeling.Fit`, optional
        Fit instance specifying the backend and fit options. If None, the `~gammapy.modeling.Fit` instance is created
        internally. Default is None.
    reoptimize : bool, optional
        If True, the free parameters of the other models are fitted in each bin independently,
        together with the norm of the source of interest
        (but the other parameters of the source of interest are kept frozen).
        If False, only the norm of the source of interest is fitted,
        and all other parameters are frozen at their current values.
        Default is False.
    sum_over_energy_groups : bool, optional
        Whether to sum over the energy groups or fit the norm on the full energy grid. Default is None.
    n_jobs : int, optional
        Number of processes used in parallel for the computation. The number of jobs is limited to the number of
        physical CPUs. If None, defaults to `~gammapy.utils.parallel.N_JOBS_DEFAULT`.
        Default is None.
    parallel_backend : {"multiprocessing", "ray"}, optional
        Which backend to use for multiprocessing. If None, defaults to `~gammapy.utils.parallel.BACKEND_DEFAULT`.
    norm : `~gammapy.modeling.Parameter` or dict, optional
        Norm parameter used for the fit.
        Default is None and a new parameter is created automatically, with value=1, name="norm",
        scan_min=0.2, scan_max=5, and scan_n_values = 11. By default, the min and max are not set
        (consider setting them if errors or upper limits computation fails). If a dict is given,
        the entries should be a subset of `~gammapy.modeling.Parameter` arguments.

    Notes
    -----
    - For further explanation, see :ref:`estimators`.
    - In case of failure of upper limits computation (e.g. nan), see the User Guide: :ref:`dropdown-UL`.

    Examples
    --------
    .. testcode::

        from astropy import units as u
        from gammapy.datasets import SpectrumDatasetOnOff
        from gammapy.estimators import FluxPointsEstimator
        from gammapy.modeling.models import PowerLawSpectralModel, SkyModel

        path = "$GAMMAPY_DATA/joint-crab/spectra/hess/"
        dataset = SpectrumDatasetOnOff.read(path + "pha_obs23523.fits")

        pwl = PowerLawSpectralModel(index=2.7, amplitude='3e-11  cm-2 s-1 TeV-1')

        dataset.models = SkyModel(spectral_model=pwl, name="crab")

        estimator = FluxPointsEstimator(
            source="crab",
            energy_edges=[0.1, 0.3, 1, 3, 10, 30, 100] * u.TeV,
        )

        fp = estimator.run(dataset)
        print(fp)

    .. testoutput::

        FluxPoints
        ----------

          geom                   : RegionGeom
          axes                   : ['lon', 'lat', 'energy']
          shape                  : (1, 1, 6)
          quantities             : ['norm', 'norm_err', 'ts', 'npred', 'npred_excess', 'stat', 'stat_null', 'counts', 'success']
          ref. model             : pl
          n_sigma                : 1
          n_sigma_ul             : 2
          sqrt_ts_threshold_ul   : 2
          sed type init          : likelihood
    """

    tag = "FluxPointsEstimator"

    def __init__(
        self,
        energy_edges=[1, 10] * u.TeV,
        sum_over_energy_groups=False,
        n_jobs=None,
        parallel_backend=None,
        **kwargs,
    ):
        self.energy_edges = energy_edges
        self.sum_over_energy_groups = sum_over_energy_groups
        self.n_jobs = n_jobs
        self.parallel_backend = parallel_backend

        fit = Fit(confidence_opts={"backend": "scipy"})
        kwargs.setdefault("fit", fit)
        super().__init__(**kwargs)

    def run(self, datasets):
        """Run the flux point estimator for all energy groups.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            Datasets.

        Returns
        -------
        flux_points : `FluxPoints`
            Estimated flux points.
        """
        if not isinstance(datasets, DatasetsActor):
            datasets = Datasets(datasets=datasets)

        if not datasets.energy_axes_are_aligned:
            raise ValueError("All datasets must have aligned energy axes.")

        telescopes = []
        for d in datasets:
            if d.meta_table is not None and "TELESCOP" in d.meta_table.colnames:
                telescopes.extend(list(d.meta_table["TELESCOP"].flatten()))
        if len(np.unique(telescopes)) > 1:
            raise ValueError(
                "All datasets must use the same value of the"
                " 'TELESCOP' meta keyword."
            )

        meta = {
            "n_sigma": self.n_sigma,
            "n_sigma_ul": self.n_sigma_ul,
            "sed_type_init": "likelihood",
        }

        rows = parallel.run_multiprocessing(
            self.estimate_flux_point,
            zip(
                repeat(datasets),
                self.energy_edges[:-1],
                self.energy_edges[1:],
            ),
            backend=self.parallel_backend,
            pool_kwargs=dict(processes=self.n_jobs),
            task_name="Energy bins",
        )

        table = Table(rows, meta=meta)
        model = _get_reference_model(datasets.models[self.source], self.energy_edges)
        return FluxPoints.from_table(
            table=table,
            reference_model=model.copy(),
            gti=datasets.gti,
            format="gadf-sed",
        )

    def estimate_flux_point(self, datasets, energy_min, energy_max):
        """Estimate flux point for a single energy group.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            Datasets.
        energy_min, energy_max : `~astropy.units.Quantity`
            Energy bounds to compute the flux point for.

        Returns
        -------
        result : dict
            Dictionary with results for the flux point.
        """
        datasets_sliced = datasets.slice_by_energy(
            energy_min=energy_min, energy_max=energy_max
        )
        if self.sum_over_energy_groups:
            datasets_sliced = datasets_sliced.__class__(
                [_.to_image(name=_.name) for _ in datasets_sliced]
            )

        if len(datasets_sliced) > 0:
            if datasets.models is not None:
                models_sliced = datasets.models._slice_by_energy(
                    energy_min=energy_min,
                    energy_max=energy_max,
                    sum_over_energy_groups=self.sum_over_energy_groups,
                )
                datasets_sliced.models = models_sliced

            return super().run(datasets=datasets_sliced)
        else:
            log.warning(f"No dataset contribute in range {energy_min}-{energy_max}")
            model = _get_reference_model(
                datasets.models[self.source], self.energy_edges
            )
            return self._nan_result(datasets, model, energy_min, energy_max)

    def _nan_result(self, datasets, model, energy_min, energy_max):
        """NaN result."""
        energy_axis = MapAxis.from_energy_edges([energy_min, energy_max])

        with np.errstate(invalid="ignore", divide="ignore"):
            result = model.reference_fluxes(energy_axis=energy_axis)
            # convert to scalar values
            result = {key: value.item() for key, value in result.items()}

        result.update(
            {
                "norm": np.nan,
                "stat": np.nan,
                "success": False,
                "norm_err": np.nan,
                "ts": np.nan,
                "counts": np.zeros(len(datasets)),
                "npred": np.nan * np.zeros(len(datasets)),
                "npred_excess": np.nan * np.zeros(len(datasets)),
                "datasets": datasets.names,
            }
        )

        if "errn-errp" in self.selection_optional:
            result.update({"norm_errp": np.nan, "norm_errn": np.nan})

        if "ul" in self.selection_optional:
            result.update({"norm_ul": np.nan})

        if "scan" in self.selection_optional:
            norm_scan = self.norm.copy().scan_values
            result.update({"norm_scan": norm_scan, "stat_scan": np.nan * norm_scan})

        if "sensitivity" in self.selection_optional:
            result.update({"norm_sensitivity": np.nan})

        return result



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst
#
# Gammapy documentation ordering configuration file.
#
# GalleryExplicitOrder is a re-implementation of private sphinx_gallery class `_SortKey`.
#


TUTORIAL_SORT_DICT = {
    #  **Tutorials**
    #  introductory
        "overview.py": 0,
        "analysis_1.py": 1,
        "analysis_2.py": 2,
    #  data exploration
        "hess.py": 0,
        "cta.py": 1,
        "magic.py": 2,
        "fermi_lat.py": 3,
        "hawc.py": 4,
        "veritas.py": 5,
    # details
        "irfs.py": 0,
        "observation_clustering.py": 1,
        "theta_square_plot.py": 2,
        "makers.py": 3,
        "datasets.py": 4,
        "maps.py": 5,
        "mask_maps.py": 6,
        "catalog.py": 7,
        "models.py": 8,
        "priors.py": 9,
        "model_management.py": 10,
        "fitting.py": 11,
        "nested_sampling_Crab.py": 12,
        "estimators.py": 13,
        "parameter_ul.py": 14,
    #  data analysis
    #  1d
        "cta_sensitivity.py": 0,
        "spectral_analysis.py": 1,
        "spectral_analysis_hli.py": 2,
        "extended_source_spectral_analysis.py": 3,
        "spectrum_simulation.py": 4,
        "sed_fitting.py": 5,
    #  2d
        "detect.py": 0,
        "ring_background.py": 1,
        "modeling_2D.py": 2,
    # 3d
        "analysis_3d.py": 0,
        "cta_data_analysis.py": 1,
        "analysis_mwl.py": 2,
        "simulate_3d.py": 3,
        "event_sampling.py": 4,
        "event_sampling_nrg_depend_models.py": 5,
        "flux_profiles.py": 6,
        "non_detected_source.py": 7,
    # time
        "light_curve.py": 0,
        "light_curve_flare.py": 1,
        "variability_estimation.py": 2,
        "time_resolved_spectroscopy.py": 3,
        "light_curve_simulation.py": 4,
    # astro
        "ebl.py": 0,
        "energy_dependent_estimation.py": 1,
        "pulsar_analysis.py": 2,
        "astro_dark_matter.py": 3,
    # scripts
        "survey_map.py": 0,
}


class BaseExplicitOrder:
    """
    Base class inspired by sphinx_gallery _SortKey to customize sorting based on a
    dictionary. The dictionary should contain the filename and its order in the
    subsection.
    """

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def __init__(self, src_dir):
        self.src_dir = src_dir


class TutorialExplicitOrder(BaseExplicitOrder):
    """
    Class that handle the ordering of the tutorials in each gallery subsection.
    """

    sort_dict = TUTORIAL_SORT_DICT

    def __call__(self, filename):
        if filename in self.sort_dict.keys():
            return self.sort_dict.get(filename)
        else:
            return 0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Pulsar velocity distribution models."""

import numpy as np
from astropy.modeling import Fittable1DModel, Parameter
from astropy.units import Quantity

__all__ = [
    "FaucherKaspi2006VelocityBimodal",
    "FaucherKaspi2006VelocityMaxwellian",
    "Paczynski1990Velocity",
    "velocity_distributions",
]

# Simulation range used for random number drawing
VMIN, VMAX = Quantity([0, 4000], "km/s")


class FaucherKaspi2006VelocityMaxwellian(Fittable1DModel):
    r"""Maxwellian pulsar velocity distribution.

    .. math::
        f(v) = A \sqrt{ \frac{2}{\pi}} \frac{v ^ 2}{\sigma ^ 3 }
               \exp \left(-\frac{v ^ 2}{2 \sigma ^ 2} \right)

    Parameters
    ----------
    amplitude : float
        Value of the integral.
    sigma : float
        Velocity parameter (km s^-1).

    References
    ----------
    .. [1] `Faucher-Giguère and Kaspi (2006), "Birth and Evolution of Isolated Radio Pulsars"
       <https://ui.adsabs.harvard.edu/abs/2006ApJ...643..332F>`_
    """

    amplitude = Parameter()
    sigma = Parameter()

    def __init__(self, amplitude=1, sigma=265, **kwargs):
        super().__init__(amplitude=amplitude, sigma=sigma, **kwargs)

    @staticmethod
    def evaluate(v, amplitude, sigma):
        """One dimensional velocity model function."""
        term1 = np.sqrt(2 / np.pi) * v**2 / sigma**3
        term2 = np.exp(-(v**2) / (2 * sigma**2))
        return term1 * term2


class FaucherKaspi2006VelocityBimodal(Fittable1DModel):
    r"""Bimodal pulsar velocity distribution.

    .. math::
        f(v) = A\sqrt{\frac{2}{\pi}} v^2 \left[\frac{w}{\sigma_1^3}
        \exp \left(-\frac{v^2}{2\sigma_1^2} \right) + \frac{1-w}{\sigma_2^3}
        \exp \left(-\frac{v^2}{2\sigma_2^2} \right) \right]

    Formula (7) in [1]_.

    Parameters
    ----------
    amplitude : float
        Value of the integral.
    sigma1 : float
        See model formula.
    sigma2 : float
        See model formula.
    w : float
        See model formula.

    References
    ----------
    .. [1] `Faucher-Giguère and Kaspi (2006), "Birth and Evolution of Isolated Radio Pulsars"
       <https://ui.adsabs.harvard.edu/abs/2006ApJ...643..332F>`_
    """

    amplitude = Parameter()
    sigma_1 = Parameter()
    sigma_2 = Parameter()
    w = Parameter()

    def __init__(self, amplitude=1, sigma_1=160, sigma_2=780, w=0.9, **kwargs):
        super().__init__(
            amplitude=amplitude, sigma_1=sigma_1, sigma_2=sigma_2, w=w, **kwargs
        )

    @staticmethod
    def evaluate(v, amplitude, sigma_1, sigma_2, w):
        """One dimensional Faucher-Guigere & Kaspi 2006 velocity model function."""
        A = amplitude * np.sqrt(2 / np.pi) * v**2
        term1 = (w / sigma_1**3) * np.exp(-(v**2) / (2 * sigma_1**2))
        term2 = (1 - w) / sigma_2**3 * np.exp(-(v**2) / (2 * sigma_2**2))
        return A * (term1 + term2)


class Paczynski1990Velocity(Fittable1DModel):
    r"""Distribution by Lyne 1982 and adopted by Paczynski and Faucher.

    .. math::
        f(v) = A\frac{4}{\pi} \frac{1}{v_0 \left[1 + (v / v_0) ^ 2 \right] ^ 2}

    Formula (3) in [1]_.

    Parameters
    ----------
    amplitude : float
        Value of the integral.
    v_0 : float
        Velocity parameter (km s^-1).

    References
    ----------
    .. [1] `Paczynski (1990), "A Test of the Galactic Origin of Gamma-Ray Bursts"
       <https://ui.adsabs.harvard.edu/abs/1990ApJ...348..485P>`_
    """

    amplitude = Parameter()
    v_0 = Parameter()

    def __init__(self, amplitude=1, v_0=560, **kwargs):
        super().__init__(amplitude=amplitude, v_0=v_0, **kwargs)

    @staticmethod
    def evaluate(v, amplitude, v_0):
        """One dimensional Paczynski 1990 velocity model function."""
        return amplitude * 4.0 / (np.pi * v_0 * (1 + (v / v_0) ** 2) ** 2)


"""Velocity distributions (dict mapping names to classes)."""
velocity_distributions = {
    "H05": FaucherKaspi2006VelocityMaxwellian,
    "F06B": FaucherKaspi2006VelocityBimodal,
    "F06P": Paczynski1990Velocity,
}



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Time related utility functions."""
import numpy as np
import astropy.units as u
from astropy.time import Time, TimeDelta

__all__ = [
    "absolute_time",
    "time_ref_from_dict",
    "time_ref_to_dict",
    "time_relative_to_ref",
]

TIME_KEYWORDS = ["MJDREFI", "MJDREFF", "TIMEUNIT", "TIMESYS", "TIMEREF"]

# TODO: implement and document this properly.
# see https://github.com/gammapy/gammapy/issues/284
TIME_REF_FERMI = Time("2001-01-01T00:00:00")

# Default time ref used for GTIs
TIME_REF_DEFAULT = Time("2000-01-01T00:00:00", scale="tt")

#: Default epoch gammapy uses for FITS files (MJDREF)
#: 0 MJD, TT
DEFAULT_EPOCH = Time(0, format="mjd", scale="tt")


def time_to_fits(time, epoch=None, unit=u.s):
    """Convert time to fits format.

    Times are stored as duration since an epoch in FITS.


    Parameters
    ----------
    time : `~astropy.time.Time`
        Time to be converted.
    epoch : `~astropy.time.Time`, optional
        Epoch to use for the time. The corresponding keywords must
        be stored in the same FITS header.
        Default is None, so the `DEFAULT_EPOCH` is used.
    unit : `~astropy.units.Unit`, optional
        Unit, should be stored as `TIMEUNIT` in the same FITS header.
        Default is u.s.

    Returns
    -------
    time : astropy.units.Quantity
        Duration since epoch.
    """
    if epoch is None:
        epoch = DEFAULT_EPOCH
    return (time - epoch).to(unit)


def time_to_fits_header(time, epoch=None, unit=u.s):
    """Convert time to fits header format.

    Times are stored as duration since an epoch in FITS.


    Parameters
    ----------
    time : `~astropy.time.Time`
        Time to be converted.
    epoch : `~astropy.time.Time`, optional
        Epoch to use for the time. The corresponding keywords must
        be stored in the same FITS header.
        Default is None, so `DEFAULT_EPOCH` is used.
    unit : `~astropy.units.Unit`, optional
        Unit, should be stored as `TIMEUNIT` in the same FITS header.
        Default is u.s.

    Returns
    -------
    header_entry : tuple(float, string)
        A float / comment tuple with the time and unit.
    """
    if epoch is None:
        epoch = DEFAULT_EPOCH
    time = time_to_fits(time, epoch)
    return time.to_value(unit), unit.to_string("fits")


def time_ref_from_dict(meta, format="mjd", scale="tt"):
    """Calculate the time reference from metadata.

    Parameters
    ----------
    meta : dict
        FITS time standard header information.
    format: str, optional
        Format of the `~astropy.time.Time` information. Default is 'mjd'.
    scale: str, optional
        Scale of the `~astropy.time.Time` information. Default is 'tt'.

    Returns
    -------
    time : `~astropy.time.Time`
        Time object with ``format='MJD'``.
    """
    scale = meta.get("TIMESYS", scale).lower()

    # some files seem to have MJDREFF as string, not as float
    mjdrefi = float(meta["MJDREFI"])
    mjdreff = float(meta["MJDREFF"])
    return Time(mjdrefi, mjdreff, format=format, scale=scale)


def time_ref_to_dict(time=None, scale="tt"):
    """Convert the epoch to the relevant FITS header keywords.

    Parameters
    ----------
    time : `~astropy.time.Time`, optional
        The reference epoch for storing time in FITS.
        Default is None, so 'DEFAULT_EPOCH' is used.
    scale: str, optional
        Scale of the `~astropy.time.Time` information.
        Default is "tt".

    Returns
    -------
    meta : dict
        FITS time standard header information.
    """
    if time is None:
        time = DEFAULT_EPOCH
    mjd = Time(time, scale=scale).mjd
    i = np.floor(mjd).astype(np.int64)
    f = mjd - i
    return dict(MJDREFI=i, MJDREFF=f, TIMESYS=scale)


def time_relative_to_ref(time, meta):
    """Convert a time using an existing reference.

    The time reference is built as MJDREFI + MJDREFF in units of MJD.
    The time will be converted to seconds after the reference.

    Parameters
    ----------
    time : `~astropy.time.Time`
        Time to be converted.
    meta : dict
        Dictionary with the keywords ``MJDREFI`` and ``MJDREFF``.

    Returns
    -------
    time_delta : `~astropy.time.TimeDelta`
        Time in seconds after the reference.
    """
    time_ref = time_ref_from_dict(meta)
    return TimeDelta(time - time_ref, format="sec")


def absolute_time(time_delta, meta):
    """Convert a MET into human-readable date and time.

    Parameters
    ----------
    time_delta : `~astropy.time.TimeDelta`
        Time in seconds after the MET reference.
    meta : dict
        Dictionary with the keywords ``MJDREFI`` and ``MJDREFF``.

    Returns
    -------
    time : `~astropy.time.Time`
        Absolute time with ``format='ISOT'`` and ``scale='UTC'``.
    """
    time = time_ref_from_dict(meta) + time_delta
    return Time(time.utc.isot)


def extract_time_info(row):
    """Extract the timing metadata from an event file header.

    Parameters
    ----------
    row : dict
        Dictionary with all the metadata of an event file.

    Returns
    -------
    time_row : dict
        Dictionary with only the time metadata.
    """
    time_row = {}
    for name in TIME_KEYWORDS:
        time_row[name] = row[name]
    return time_row


def unique_time_info(rows):
    """Check if the time information are identical between all metadata dictionaries.

    Parameters
    ----------
    rows : list
        List of dictionaries with a list of time metadata from different observations.

    Returns
    -------
    status : bool
        True if the time metadata are identical between observations.
    """
    if len(rows) <= 1:
        return True

    first_obs = rows[0]
    for row in rows[1:]:
        for name in TIME_KEYWORDS:
            if first_obs[name] != row[name] or row[name] is None:
                return False
    return True



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
from astropy import units as u
from gammapy.irf import PSF3D
from gammapy.maps import MapAxis
from gammapy.utils.testing import mpl_plot_check, requires_data


@pytest.fixture(scope="session")
def psf_3d():
    filename = "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_023523.fits.gz"
    return PSF3D.read(filename, hdu="PSF")


def test_psf_3d_wrong_units():
    energy_axis = MapAxis.from_energy_edges([80, 125] * u.GeV, name="energy_true")

    offset_axis = MapAxis.from_edges([0, 1, 2], unit="deg", name="offset")

    rad_axis = MapAxis.from_edges([0, 1, 2], unit="deg", name="rad")

    wrong_unit = u.cm**2 * u.s
    data = np.ones((energy_axis.nbin, offset_axis.nbin, rad_axis.nbin)) * wrong_unit
    psf3d_test = PSF3D(axes=[energy_axis, offset_axis, rad_axis])
    with pytest.raises(ValueError) as error:
        PSF3D(axes=[energy_axis, offset_axis, rad_axis], data=data)
        assert error.match(
            f"Error: {wrong_unit} is not an allowed unit. {psf3d_test.tag} requires "
            f"{psf3d_test.default_unit} data quantities."
        )


@requires_data()
def test_psf_3d_basics(psf_3d):
    rad_axis = psf_3d.axes["rad"]
    assert_allclose(rad_axis.edges[-2].value, 0.659048, rtol=1e-5)
    assert rad_axis.nbin == 144
    assert rad_axis.unit == "deg"

    energy_axis_true = psf_3d.axes["energy_true"]
    assert_allclose(energy_axis_true.edges[0].value, 0.01)
    assert energy_axis_true.nbin == 32
    assert energy_axis_true.unit == "TeV"

    assert psf_3d.data.shape == (32, 6, 144)
    assert psf_3d.unit == "sr-1"

    psf_3d_new_unit = psf_3d.to_unit("deg-2")
    assert_allclose(psf_3d_new_unit.data, psf_3d.data / 3282.8063, rtol=1e-6)

    assert_allclose(psf_3d.meta["LO_THRES"], 0.01)

    assert "PSF3D" in str(psf_3d)

    with pytest.raises(ValueError):
        PSF3D(axes=psf_3d.axes, data=psf_3d.data.T)


@requires_data()
def test_psf_3d_evaluate(psf_3d):
    q = psf_3d.evaluate(energy_true="1 TeV", offset="0.3 deg", rad="0.1 deg")
    assert_allclose(q.value, 25847.249548)
    # TODO: is this the shape we want here?
    assert q.shape == ()
    assert q.unit == "sr-1"


@requires_data()
def test_psf_3d_containment_radius(psf_3d):
    q = psf_3d.containment_radius(
        energy_true=1 * u.TeV, fraction=0.68, offset=0 * u.deg
    )
    assert_allclose(q.value, 0.123352, rtol=1e-2)
    assert q.isscalar
    assert q.unit == "deg"

    q = psf_3d.containment_radius(
        energy_true=[1, 3] * u.TeV, fraction=0.68, offset=0 * u.deg
    )
    assert_allclose(q.value, [0.123261, 0.13131], rtol=1e-2)
    assert q.shape == (2,)


@requires_data()
def test_psf_3d_plot_vs_rad(psf_3d):
    with mpl_plot_check():
        psf_3d.plot_psf_vs_rad()


@requires_data()
def test_psf_3d_plot_containment(psf_3d):
    with mpl_plot_check():
        psf_3d.plot_containment_radius()


@requires_data()
def test_psf_3d_peek(psf_3d):
    with mpl_plot_check():
        psf_3d.peek()



"""
Constraining parameter limits
=============================

Explore how to deal with upper limits on parameters.

Prerequisites
-------------

It is advisable to understand the general Gammapy modelling and fitting framework before proceeding
with this notebook, e.g. see :doc:`/user-guide/modeling`.

Context
-------

Even with significant detection of a source, constraining specific model parameters may remain difficult,
allowing only for the calculation of confidence intervals.

Proposed approach
-----------------

In this section, we will use 6 observations of the blazar PKS 2155-304, taken in 2008 by
H.E.S.S, to constrain the curvature in the spectrum.

"""

######################################################################
# Setup
# -----
#
# As usual, let’s start with some general imports…
#

# %matplotlib inline
import matplotlib.pyplot as plt

import numpy as np
import astropy.units as u
from gammapy.datasets import SpectrumDatasetOnOff, Datasets
from gammapy.modeling import Fit, select_nested_models
from gammapy.modeling.models import SkyModel, LogParabolaSpectralModel
from gammapy.estimators import FluxPointsEstimator

######################################################################
# Load observation
# ----------------
#
# We will use a `~gammapy.datasets.SpectrumDatasetOnOff` to see how to constrain
# model parameters. This dataset was obtained from H.E.S.S. observation of the blazar PKS 2155-304.
# Detailed modeling of this dataset can be found in the
# :doc:`/tutorials/astrophysics/ebl` notebook.
#

dataset_onoff = SpectrumDatasetOnOff.read(
    "$GAMMAPY_DATA/PKS2155-steady/pks2155-304_steady.fits.gz"
)
dataset_onoff.peek()
plt.show()


######################################################################
# Fit spectrum
# ------------
#
# We will investigate the presence of spectral curvature by modeling the
# observed spectrum using a `~gammapy.modeling.models.LogParabolaSpectralModel`.
#

spectral_model = LogParabolaSpectralModel(
    amplitude="5e-12 TeV-1 s-1 cm-2", alpha=2, beta=0.5, reference=1.0 * u.TeV
)

model_pks = SkyModel(spectral_model, name="model_pks")
dataset_onoff.models = model_pks

fit = Fit()
result_pks = fit.run(dataset_onoff)
print(result_pks.models)


######################################################################
# We see that the parameter ``beta`` (the curvature parameter) is poorly
# constrained as the errors are very large.
# Therefore, we will perform a likelihood ratio test to evaluate the significance
# of the curvature compared to the null hypothesis of no curvature. In the null
# hypothesis, ``beta=0``.

LLR = select_nested_models(
    datasets=Datasets(dataset_onoff),
    parameters=[model_pks.parameters["beta"]],
    null_values=[0],
)
print(LLR)


######################################################################
# We can see that the improvement in the test statistic after including the
# curvature is only ~0.3, which corresponds to a significance of only 0.5.
#
# We can safely conclude that the addition of the curvature parameter does
# not significantly improve the fit. As a result, the function has internally updated
# the best fit model to the one corresponding to the null hypothesis (i.e. ``beta=0``).

print(dataset_onoff.models)


######################################################################
# Compute parameter asymmetric errors and upper limits
# ----------------------------------------------------
# In such a case, it can still be useful to be able to constrain
# the allowed range of the non-significant parameter (e.g.: to rule
# out parameter values, to compare from theoretical predications, etc.).
#
# First, we reset the alternative model on the dataset:

dataset_onoff.models = LLR["fit_results"].models
parameter = dataset_onoff.models.parameters["beta"]

######################################################################
# We can then compute the asymmetric errors and upper limits on the parameter
# of interest. It is always useful to ensure that the fit the converged by looking at the
# ``success`` and ``message`` keywords.

res_1sig = fit.confidence(datasets=dataset_onoff, parameter=parameter, sigma=1)
print(res_1sig)

######################################################################
# We can directly use this to compute :math:`n\sigma` upper limits on the parameter:

res_2sig = fit.confidence(datasets=dataset_onoff, parameter=parameter, sigma=2)
ll_2sigma = parameter.value - res_2sig["errn"]
ul_2sigma = parameter.value + res_2sig["errp"]

print(f"2-sigma lower limit on beta is {ll_2sigma:.2f}")
print(f"2-sigma upper limit on beta is {ul_2sigma:.2f}")

######################################################################
# Likelihood profile
# ------------------
# We can also compute the likelihood profile of the parameter.
# First we define the scan range such that it encompasses more than the 2-sigma parameter limits.
# Then we call `~gammapy.modeling.Fit.stat_profile` :

parameter.scan_n_values = 25
parameter.scan_min = parameter.value - 2.5 * res_2sig["errn"]
parameter.scan_max = parameter.value + 2.5 * res_2sig["errp"]
parameter.interp = "lin"
profile = fit.stat_profile(datasets=dataset_onoff, parameter=parameter, reoptimize=True)

######################################################################
# The resulting `profile` is a dictionary that stores the likelihood value and the fit result
# for each value of beta.

print(profile)


######################################################################
# Let's plot everything together

values = profile["model_pks.spectral.beta_scan"]
loglike = profile["stat_scan"]
ax = plt.gca()
ax.plot(values, loglike - np.min(loglike))
ax.set_xlabel("Beta")
ax.set_ylabel(r"$\Delta$TS")
ax.set_title(r"$\beta$-parameter likelihood profile")
ax.fill_betweenx(
    x1=parameter.value - res_2sig["errn"],
    x2=parameter.value + res_2sig["errp"],
    y=[-0.5, 25],
    alpha=0.3,
    color="pink",
    label="1-sigma range",
)
ax.fill_betweenx(
    x1=parameter.value - res_1sig["errn"],
    x2=parameter.value + res_1sig["errp"],
    y=[-0.5, 25],
    alpha=0.3,
    color="salmon",
    label="2-sigma range",
)
ax.set_ylim(-0.5, 25)
plt.legend()
plt.show()
# sphinx_gallery_thumbnail_number = 2


######################################################################
# Impact of the model choice on the flux upper limits
# ---------------------------------------------------
# The flux points depends on the underlying model assumption.
# This can have a non-negligible impact on the flux upper limits in the energy range
# where the model is not well constrained as illustrated in the following figure.
# So quote preferably upper limits from the model which is the most supported by the data.

energies = dataset_onoff.geoms["geom"].axes["energy"].edges
fpe = FluxPointsEstimator(energy_edges=energies, n_jobs=4, selection_optional=["ul"])

# Null hypothesis -- no curvature
dataset_onoff.models = LLR["fit_results_null"].models
fp_null = fpe.run(dataset_onoff)

# Alternative hypothesis -- with curvature
dataset_onoff.models = LLR["fit_results"].models
fp_alt = fpe.run(dataset_onoff)

#####################################################################
# Plot them together

ax = fp_null.plot(sed_type="e2dnde", color="blue")
LLR["fit_results_null"].models[0].spectral_model.plot(
    ax=ax,
    energy_bounds=(energies[0], energies[-1]),
    sed_type="e2dnde",
    color="blue",
    label="No curvature",
)
LLR["fit_results_null"].models[0].spectral_model.plot_error(
    ax=ax,
    energy_bounds=(energies[0], energies[-1]),
    sed_type="e2dnde",
    facecolor="blue",
    alpha=0.2,
)


fp_alt.plot(ax=ax, sed_type="e2dnde", color="red")
LLR["fit_results"].models[0].spectral_model.plot(
    ax=ax,
    energy_bounds=(energies[0], energies[-1]),
    sed_type="e2dnde",
    color="red",
    label="with curvature",
)
LLR["fit_results"].models[0].spectral_model.plot_error(
    ax=ax,
    energy_bounds=(energies[0], energies[-1]),
    sed_type="e2dnde",
    facecolor="red",
    alpha=0.2,
)

plt.legend()
plt.show()

######################################################################
# This logic can be extended to any spectral or spatial feature. As an
# exercise, try to compute the 95% spatial extent on the MSH 15-52 dataset
# used for the ring background notebook.
#



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import html
import numpy as np
import astropy.units as u
import matplotlib.pyplot as plt
from gammapy.maps import Map
from gammapy.modeling.models import PowerLawSpectralModel

__all__ = ["PSFKernel"]


class PSFKernel:
    """PSF kernel for `~gammapy.maps.Map`.

    This is a container class to store a PSF kernel
    that can be used to convolve `~gammapy.maps.WcsNDMap` objects.
    It is usually computed from an `~gammapy.irf.PSFMap`.

    Parameters
    ----------
    psf_kernel_map : `~gammapy.maps.Map`
        PSF kernel stored in a Map.
    normalize : bool, optional
        Normalize the kernel. Default is True.

    Examples
    --------
    .. testcode::

        from gammapy.maps import Map, WcsGeom, MapAxis
        from gammapy.irf import PSFMap
        from astropy import units as u

        # Define energy axis
        energy_axis_true = MapAxis.from_energy_bounds(
            "0.1 TeV", "10 TeV", nbin=3, name="energy_true"
        )

        # Create WcsGeom and map
        geom = WcsGeom.create(binsz=0.02 * u.deg, width=2.0 * u.deg, axes=[energy_axis_true])
        some_map = Map.from_geom(geom)

        # Fill map at three positions
        some_map.fill_by_coord([[0, 0, 0], [0, 0, 0], [0.3, 1, 3]])

        psf = PSFMap.from_gauss(
            energy_axis_true=energy_axis_true, sigma=[0.3, 0.2, 0.1] * u.deg
        )

        kernel = psf.get_psf_kernel(geom=geom, max_radius=1*u.deg)

        # Do the convolution
        some_map_convolved = some_map.convolve(kernel)

        some_map_convolved.plot_grid() # doctest: +SKIP
    """

    def __init__(self, psf_kernel_map, normalize=True):
        self._psf_kernel_map = psf_kernel_map

        if normalize:
            self.normalize()

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    def normalize(self):
        """Force normalisation of the kernel."""
        data = self.psf_kernel_map.data
        if self.psf_kernel_map.geom.is_image:
            axis = (0, 1)
        else:
            axis = (1, 2)

        with np.errstate(divide="ignore", invalid="ignore"):
            data = np.nan_to_num(data / data.sum(axis=axis, keepdims=True))
            self.psf_kernel_map.data = data

    @property
    def data(self):
        """Access the PSFKernel numpy array."""
        return self._psf_kernel_map.data

    @property
    def psf_kernel_map(self):
        """The map object holding the kernel as a `~gammapy.maps.Map`."""
        return self._psf_kernel_map

    @classmethod
    def read(cls, *args, **kwargs):
        """Read kernel Map from file."""
        psf_kernel_map = Map.read(*args, **kwargs)
        return cls(psf_kernel_map)

    @classmethod
    def from_spatial_model(cls, model, geom, max_radius=None, factor=4):
        """Create PSF kernel from spatial model.

        Parameters
        ----------
        geom : `~gammapy.maps.WcsGeom`
            Map geometry.
        model : `~gammapy.modeling.models.SpatiaModel`
            Gaussian width.
        max_radius : `~astropy.coordinates.Angle`, optional
            Desired kernel map size. Default is None.
        factor : int, optional
            Oversample factor to compute the PSF. Default is 4.

        Returns
        -------
        kernel : `~gammapy.irf.PSFKernel`
            The kernel Map with reduced geometry according to the max_radius.
        """
        if max_radius is None:
            max_radius = model.evaluation_radius

        geom = geom.to_odd_npix(max_radius=max_radius)
        model.position = geom.center_skydir

        geom = geom.upsample(factor=factor)
        map = model.integrate_geom(geom, oversampling_factor=1)
        return cls(psf_kernel_map=map.downsample(factor=factor))

    @classmethod
    def from_gauss(cls, geom, sigma, max_radius=None, factor=4):
        """Create Gaussian PSF.

        This is used for testing and examples.
        The map geometry parameters (pixel size, energy bins) are taken from ``geom``.
        The Gaussian width ``sigma`` is a scalar.

        Parameters
        ----------
        geom : `~gammapy.maps.WcsGeom`
            Map geometry.
        sigma : `~astropy.coordinates.Angle`
            Gaussian width.
        max_radius : `~astropy.coordinates.Angle`, optional
            Desired kernel map size. Default is None.
        factor : int, optional
            Oversample factor to compute the PSF. Default is 4.

        Returns
        -------
        kernel : `~gammapy.irf.PSFKernel`
            The kernel Map with reduced geometry according to the max_radius.
        """
        # TODO : support array input if it should vary along the energy axis.
        from gammapy.modeling.models import GaussianSpatialModel

        gauss = GaussianSpatialModel(sigma=sigma)

        return cls.from_spatial_model(
            model=gauss, geom=geom, max_radius=max_radius, factor=factor
        )

    def write(self, *args, **kwargs):
        """Write the Map object which contains the PSF kernel to file."""
        self.psf_kernel_map.write(*args, **kwargs)

    def to_image(self, spectral_model=None, exposure=None, keepdims=True):
        """Transform 3D PSFKernel into a 2D PSFKernel.

        Parameters
        ----------
        spectral_model : `~gammapy.modeling.models.SpectralModel`, optional
            Spectral model to compute the weights.
            Default is power-law with spectral index of 2.
        exposure : `~astropy.units.Quantity` or `~numpy.ndarray`, optional
            1D array containing exposure in each true energy bin.
            It must have the same size as the PSFKernel energy axis.
            Default is uniform exposure over energy.
        keepdims : bool, optional
            If true, the resulting PSFKernel will keep an energy axis with one bin.
            Default is True.

        Returns
        -------
        weighted_kernel : `~gammapy.irf.PSFKernel`
            The weighted kernel summed over energy.
        """
        map = self.psf_kernel_map

        if spectral_model is None:
            spectral_model = PowerLawSpectralModel(index=2.0)

        if exposure is None:
            exposure = np.ones(map.geom.axes[0].center.shape)

        exposure = u.Quantity(exposure)

        if exposure.shape != map.geom.axes[0].center.shape:
            raise ValueError("Incorrect exposure_array shape")

        # Compute weights vector
        for name in map.geom.axes.names:
            if "energy" in name:
                energy_name = name
        energy_edges = map.geom.axes[energy_name].edges
        weights = spectral_model.integral(
            energy_min=energy_edges[:-1], energy_max=energy_edges[1:], intervals=True
        )
        weights *= exposure
        weights /= weights.sum()

        spectrum_weighted_kernel = map.copy()
        spectrum_weighted_kernel.quantity *= weights[:, np.newaxis, np.newaxis]

        return self.__class__(spectrum_weighted_kernel.sum_over_axes(keepdims=keepdims))

    def slice_by_idx(self, slices):
        """Slice by index."""
        kernel = self.psf_kernel_map.slice_by_idx(slices=slices)
        return self.__class__(psf_kernel_map=kernel)

    def plot_kernel(self, ax=None, energy=None, add_cbar=False, **kwargs):
        """Plot PSF kernel.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`, optional
            Matplotlib axes. Default is None.
        energy : `~astropy.units.Quantity`, optional
            If None, the PSF kernel is summed over the energy axis. Otherwise, the kernel
            corresponding to the energy bin including the given energy is shown.
        add_cbar : bool, optional
            Add a colorbar. Default is False.
        kwargs: dict
            Keyword arguments passed to `~matplotlib.axes.Axes.imshow`.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            Matplotlib axes.
        """
        ax = plt.gca() if ax is None else ax

        if energy is not None:
            kernel_map = self.psf_kernel_map
            energy_center = kernel_map.geom.axes["energy_true"].center.to(energy.unit)
            idx = np.argmin(np.abs(energy_center.value - energy.value))
            kernel_map.get_image_by_idx([idx]).plot(ax=ax, add_cbar=add_cbar, **kwargs)
        else:
            self.to_image().psf_kernel_map.plot(ax=ax, add_cbar=add_cbar, **kwargs)

        return ax

    def peek(self, figsize=(15, 5)):
        """Quick-look summary plots.

        This method creates a figure with two subplots:

        * PSF kernel plot : PSF kernel summed over the energy axis
        * PSF kernel plot : PSF kernel at 1 TeV

        Parameters
        ----------
        figsize : tuple, optional
            Size of the figure. Default is (15, 5).
        """
        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=figsize)

        axes[0].set_title("Energy-integrated PSF kernel")
        self.plot_kernel(ax=axes[0], add_cbar=True)

        axes[1].set_title("PSF kernel at 1 TeV")
        self.plot_kernel(ax=axes[1], add_cbar=True, energy=1 * u.TeV)

        plt.tight_layout()



"""
Spectral analysis of extended sources
=====================================

Perform a spectral analysis of an extended source.

Prerequisites
-------------

-  Understanding of spectral analysis techniques in classical Cherenkov
   astronomy.
-  Understanding the basic data reduction and modeling/fitting processes
   with the gammapy library API as shown in the tutorial :doc:`/tutorials/starting/analysis_2`

Context
-------

Many VHE sources in the Galaxy are extended. Studying them with a 1D
spectral analysis is more complex than studying point sources. One often
has to use complex (i.e. non-circular) regions and more importantly, one
has to take into account the fact that the instrument response is non-uniform
over the selected region. A typical example is given by the
supernova remnant RX J1713-3935 which is nearly 1 degree in diameter.
See the `following
article <https://ui.adsabs.harvard.edu/abs/2018A%26A...612A...6H/abstract>`__.

**Objective: Measure the spectrum of RX J1713-3945 in a 1 degree region
fully enclosing it.**

Proposed approach
-----------------

We have seen in the general presentation of the spectrum extraction for
point sources (see :doc:`/tutorials/analysis-1d/spectral_analysis`
tutorial) that Gammapy uses specific
datasets makers to first produce reduced spectral data and then to
extract OFF measurements with reflected background techniques: the
`~gammapy.makers.SpectrumDatasetMaker` and the
`~gammapy.makers.ReflectedRegionsBackgroundMaker`. However, if the flag
`use_region_center` is not set to `False`, the former simply
computes the reduced IRFs at the center of the ON region (assumed to be
circular).

This is no longer valid for extended sources. To be able to compute
average responses in the ON region, we can set
`use_region_center=False` with the
`~gammapy.makers.SpectrumDatasetMaker`, in which case the values of
the IRFs are averaged over the entire region.

In summary, we have to:

-  Define an ON region (a `~regions.SkyRegion`) fully enclosing the
   source we want to study.
-  Define a `~gammapy.maps.RegionGeom` with the ON region and the
   required energy range (in particular, beware of the true energy range).
-  Create the necessary makers :

   -  the spectrum dataset maker :
      `~gammapy.makers.SpectrumDatasetMaker` with
      `use_region_center=False`
   -  the OFF background maker, here a
      `~gammapy.makers.ReflectedRegionsBackgroundMaker`
   -  and usually the safe range maker :
      `~gammapy.makers.SafeMaskMaker`

-  Perform the data reduction loop. And for every observation:

   -  Produce a spectrum dataset
   -  Extract the OFF data to produce a
      `~gammapy.datasets.SpectrumDatasetOnOff` and compute a safe
      range for it.
   -  Stack or store the resulting spectrum dataset.

-  Finally proceed with model fitting on the dataset as usual.

Here, we will use the RX J1713-3945 observations from the H.E.S.S. first
public test data release. The tutorial is implemented with the
intermediate level API.

Setup
-----

As usual, we’ll start with some general imports…

"""

import astropy.units as u
from astropy.coordinates import Angle, SkyCoord
from regions import CircleSkyRegion

# %matplotlib inline
import matplotlib.pyplot as plt
from IPython.display import display
from gammapy.data import DataStore
from gammapy.datasets import Datasets, SpectrumDataset
from gammapy.makers import (
    ReflectedRegionsBackgroundMaker,
    SafeMaskMaker,
    SpectrumDatasetMaker,
)
from gammapy.maps import MapAxis, RegionGeom
from gammapy.modeling import Fit
from gammapy.modeling.models import PowerLawSpectralModel, SkyModel


######################################################################
# Select the data
# ---------------
#
# We first set the datastore and retrieve a few observations from our
# source.
#

datastore = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1/")
obs_ids = [20326, 20327, 20349, 20350, 20396, 20397]
# In case you want to use all RX J1713 data in the H.E.S.S. DR1
# other_ids=[20421, 20422, 20517, 20518, 20519, 20521, 20898, 20899, 20900]

observations = datastore.get_observations(obs_ids)


######################################################################
# Prepare the datasets creation
# -----------------------------
#


######################################################################
# Select the ON region
# ~~~~~~~~~~~~~~~~~~~~
#
# Here we take a simple 1 degree circular region because it fits well with
# the morphology of RX J1713-3945. More complex regions could be used
# e.g. `~regions.EllipseSkyRegion` or `~regions.RectangleSkyRegion`.
#

target_position = SkyCoord(347.3, -0.5, unit="deg", frame="galactic")
radius = Angle("0.5 deg")
on_region = CircleSkyRegion(target_position, radius)


######################################################################
# Define the geometries
# ~~~~~~~~~~~~~~~~~~~~~
#
# This part is especially important.
#
# -  We have to define first energy axes. They define the axes of the resulting
#    `~gammapy.datasets.SpectrumDatasetOnOff`. In particular, we have to be
#    careful to the true energy axis: it has to cover a larger range than the
#    reconstructed energy one.
# -  Then we define the region geometry itself from the on region.
#

# The binning of the final spectrum is defined here.
energy_axis = MapAxis.from_energy_bounds(0.1, 40.0, 10, unit="TeV")

# Reduced IRFs are defined in true energy (i.e. not measured energy).
energy_axis_true = MapAxis.from_energy_bounds(
    0.05, 100, 30, unit="TeV", name="energy_true"
)

geom = RegionGeom(on_region, axes=[energy_axis])


######################################################################
# Create the makers
# ~~~~~~~~~~~~~~~~~
#
# First we instantiate the target `~gammapy.datasets.SpectrumDataset`.
#

dataset_empty = SpectrumDataset.create(
    geom=geom,
    energy_axis_true=energy_axis_true,
)


######################################################################
# Now we create its associated maker. Here we need to produce, counts,
# exposure and edisp (energy dispersion) entries. PSF and IRF background
# are not needed, therefore we don’t compute them.
#
# **IMPORTANT**: Note that `use_region_center` is set to `False`. This
# is necessary so that the `~gammapy.makers.SpectrumDatasetMaker`
# considers the whole region in the IRF computation and not only the
# center.
#

maker = SpectrumDatasetMaker(
    selection=["counts", "exposure", "edisp"], use_region_center=False
)


######################################################################
# Now we create the OFF background maker for the spectra. If we have an
# exclusion region, we have to pass it here. We also define the safe range
# maker.
#

bkg_maker = ReflectedRegionsBackgroundMaker()
safe_mask_maker = SafeMaskMaker(methods=["aeff-max"], aeff_percent=10)


######################################################################
# Perform the data reduction loop.
# --------------------------------
#
# We can now run over selected observations. For each of them, we:
#
# -   Create the `~gammapy.datasets.SpectrumDataset`
# -  Compute the OFF via the reflected background method and create a `~gammapy.datasets.SpectrumDatasetOnOff` object
# -  Run the safe mask maker on it
# -  Add the `~gammapy.datasets.SpectrumDatasetOnOff` to the list.
#

datasets = Datasets()

for obs in observations:
    # A SpectrumDataset is filled in this geometry
    dataset = maker.run(dataset_empty.copy(name=f"obs-{obs.obs_id}"), obs)

    # Define safe mask
    dataset = safe_mask_maker.run(dataset, obs)

    # Compute OFF
    dataset = bkg_maker.run(dataset, obs)

    # Append dataset to the list
    datasets.append(dataset)

display(datasets.meta_table)


######################################################################
# Explore the results
# -------------------
#
# We can peek at the content of the spectrum datasets
#

datasets[0].peek()
plt.show()


######################################################################
# Cumulative excess and significance
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Finally, we can look at cumulative significance and number of excesses.
# This is done with the `info_table` method of
# `~gammapy.datasets.Datasets`.
#

info_table = datasets.info_table(cumulative=True)

display(info_table)

######################################################################
# And make the corresponding plots

fig, (ax_excess, ax_sqrt_ts) = plt.subplots(figsize=(10, 4), ncols=2, nrows=1)
ax_excess.plot(
    info_table["livetime"].to("h"),
    info_table["excess"],
    marker="o",
    ls="none",
)
ax_excess.set_title("Excess")
ax_excess.set_xlabel("Livetime [h]")
ax_excess.set_ylabel("Excess events")

ax_sqrt_ts.plot(
    info_table["livetime"].to("h"),
    info_table["sqrt_ts"],
    marker="o",
    ls="none",
)

ax_sqrt_ts.set_title("Sqrt(TS)")
ax_sqrt_ts.set_xlabel("Livetime [h]")
ax_sqrt_ts.set_ylabel("Sqrt(TS)")
plt.show()


######################################################################
# Perform spectral model fitting
# ------------------------------
#
# Here we perform a joint fit.
#
# We first create the model, here a simple powerlaw, and assign it to
# every dataset in the `~gammapy.datasets.Datasets`.
#

spectral_model = PowerLawSpectralModel(
    index=2, amplitude=2e-11 * u.Unit("cm-2 s-1 TeV-1"), reference=1 * u.TeV
)
model = SkyModel(spectral_model=spectral_model, name="RXJ 1713")

datasets.models = [model]


######################################################################
# Now we can run the fit
#

fit_joint = Fit()
result_joint = fit_joint.run(datasets=datasets)
print(result_joint)


######################################################################
# Explore the fit results
# ~~~~~~~~~~~~~~~~~~~~~~~
#
# First the fitted parameters values and their errors.
#

display(datasets.models.to_parameters_table())


######################################################################
# Then plot the fit result to compare measured and expected counts. Rather
# than plotting them for each individual dataset, we stack all datasets
# and plot the fit result on the result.
#

# First stack them all
reduced = datasets.stack_reduce()

# Assign the fitted model
reduced.models = model

# Plot the result
ax_spectrum, ax_residuals = reduced.plot_fit()
plt.show()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Sampler parameter classes."""

from .utils import _parse_datasets

__all__ = ["Sampler", "SamplerLikelihood", "SamplerResult"]


class Sampler:
    """Sampler class.

    The sampler class provides a uniform interface to multiple sampler backends. Currently available: "UltraNest".

    Parameters
    ----------
    backend : {"ultranest"}
        Global backend used for sampler. Default is "ultranest".
        UltraNest: Most options can be found in the
        `UltraNest doc <https://johannesbuchner.github.io/UltraNest/>`__.
    sampler_opts : dict, optional
        Sampler options passed to the sampler. See the full list of options on the
        `UltraNest documentation <https://johannesbuchner.github.io/UltraNest/ultranest.html#ultranest.integrator.ReactiveNestedSampler>`__.
        Noteworthy options:

        live_points : int
            Minimum number of live points used in the sampling. Increase this number to get more accurate results.
            For more samples in the posterior increase this number or the min_ess parameter.
            Default is 400 live points.
        frac_remain : float
            Integrate until this fraction of the integral is left in the remainder. Set to a low number (1e-2 … 1e-5)
            to make sure peaks are discovered. Set to a higher number (0.5) if you know the posterior is simple.
            Default is 0.5.
        min_ess : int
            Target number of effective posterior samples. Increase this number to get more accurate results.
            Default is live_points, but you may need to increase it to 1000 or more for complex posteriors.
        log_dir : str
            Where to store output files.
            Default is None and no results are not stored.
        resume : str
            ‘overwrite’, overwrite previous data. ‘subfolder’, create a fresh subdirectory in `log_dir`.
            ‘resume’ or True, continue previous run if available. Only works when dimensionality, transform or
            likelihood are consistent.
        step_sampler : bool, optional
            Use a step sampler. This can be more efficient for higher dimensions (>10 or 15 parameters), but much
            slower for lower dimensions.
            Default is False.
        nsteps : int
            Number of steps to take in each direction in the step sampler. Increase this number to get more
            accurate results at the cost of more computation time.
            Default is 10.
    run_opts : dict, optional
        Optional run options passed to the given backend when running the sampler.
        See the full list of run options on the
        `UltraNest documentation <https://johannesbuchner.github.io/UltraNest/ultranest.html#ultranest.integrator.ReactiveNestedSampler.run>`__.

    Examples
    --------
    For a usage example, see :doc:`/tutorials/details/nested_sampling_Crab` tutorial.

    Notes
    -----
    If you are using the "UltraNest" library, please follow its citation scheme:
    `Cite UltraNest <https://johannesbuchner.github.io/UltraNest/issues.html#how-should-i-cite-ultranest>`__.

    """

    # TODO: add "zeusmc", "emcee"

    def __init__(self, backend="ultranest", sampler_opts=None, run_opts=None):
        self._sampler = None
        self.backend = backend
        self.sampler_opts = {} if sampler_opts is None else sampler_opts
        self.run_opts = {} if run_opts is None else run_opts

        if self.backend == "ultranest":
            self.sampler_opts.setdefault("live_points", 400)
            self.sampler_opts.setdefault("frac_remain", 0.5)
            self.sampler_opts.setdefault("log_dir", None)
            self.sampler_opts.setdefault("resume", "subfolder")
            self.sampler_opts.setdefault("step_sampler", False)
            self.sampler_opts.setdefault("nsteps", 10)

    @staticmethod
    def _update_models_from_posterior(models, result):
        """
        Update the models with the posterior distribution.

        But this raises the question on how to estimate the error given the median and maxLogL.
        Covariance matrix is not defined in this sample approach (could be approximated via the samples).

        Parameters
        ----------
        models : `~gammapy.modeling.models`
            The models to update
        result : dict
            The sampler results dictionary containing the posterior distribution infos.
        """
        # TODO : add option for median, distribution peak, and maxLogL once Param object has asym errors

        posterior = result["posterior"]
        for i, par in enumerate(models.parameters.free_unique_parameters):
            par.value = posterior["mean"][i]
            par.error = posterior["stdev"][i]
        models._covariance = None

    def sampler_ultranest(self, parameters, like):
        """
        Defines the Ultranest sampler and options.

        Returns the result in the SamplerResult that contains the updated models, samples, posterior distribution and
        other information.

        Parameters
        ----------
        parameters : `~gammapy.modeling.Parameters`
            The models parameters to sample.
        like : `~gammapy.modeling.sampler.SamplerLikelihood`
            The likelihood function.

        Returns
        -------
        result : `~gammapy.modeling.sampler.SamplerResult`
            The sampler results.
        """
        import ultranest

        def _prior_inverse_cdf(values):
            """Returns a list of model parameters for a given list of values (that are bound in [0,1])."""
            if None in parameters:
                raise ValueError(
                    "Some parameters have no prior set. You need priors on all parameters."
                )
            return [par.prior._inverse_cdf(val) for par, val in zip(parameters, values)]

        self._sampler = ultranest.ReactiveNestedSampler(
            parameters.names,
            like.fcn,
            transform=_prior_inverse_cdf,
            log_dir=self.sampler_opts["log_dir"],
            resume=self.sampler_opts["resume"],
        )

        if self.sampler_opts["step_sampler"]:
            from ultranest.stepsampler import (
                SliceSampler,
                generate_mixture_random_direction,
            )

            self._sampler.stepsampler = SliceSampler(
                nsteps=self.sampler_opts["nsteps"],
                generate_direction=generate_mixture_random_direction,
            )

        result = self._sampler.run(
            min_num_live_points=self.sampler_opts["live_points"],
            frac_remain=self.sampler_opts["frac_remain"],
            **self.run_opts,
        )

        return result

    def run(self, datasets):
        """
        Run the sampler on the provided datasets.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            Datasets to fit.

        Returns
        -------
        result : `~gammapy.modeling.sampler.SamplerResult`
            The sampler results. See the class description to get the exact content.
        """
        datasets, parameters = _parse_datasets(datasets=datasets)
        parameters = parameters.free_unique_parameters

        if self.backend == "ultranest":
            like = SamplerLikelihood(
                function=datasets._stat_sum_likelihood, parameters=parameters
            )
            result_dict = self.sampler_ultranest(parameters, like)
            self._sampler.print_results()

            models_copy = datasets.models.copy()
            self._update_models_from_posterior(models_copy, result_dict)

            result = SamplerResult.from_ultranest(result_dict)
            result.models = models_copy

            return result
        else:
            raise ValueError(f"Sampler {self.backend} is not supported.")


class SamplerResult:
    """SamplerResult class.

    Parameters
    ----------
    nfev : int
        Number of likelihood calls/evaluations.
    success : bool
        Did the sampler succeed in finding a good fit? Definition of convergence depends on the sampler backend.
    models : `~gammapy.modeling.models`
        The models updated after the sampler run.
    samples : `~numpy.ndarray`
        Array of (weighted) samples that can be used for histograms or corner plots.
    sampler_results : dict
        Output of sampler.
        See the :doc:`/tutorials/details/nested_sampling_Crab` tutorial for a complete description.
    """

    # TODO:
    #  - Support parameter posteriors directly on Parameter
    #    - e.g. adding a errn and errp entry
    #    - or creating a posterior entry on Parameter
    #  - Or support with a specific entry on the SamplerResult
    #  - Add write method to be consistent with FitResult

    def __init__(
        self, nfev=0, success=False, models=None, samples=None, sampler_results=None
    ):
        self.nfev = nfev
        self.success = success
        self.models = models
        self.samples = samples
        self.sampler_results = sampler_results

    @classmethod
    def from_ultranest(cls, ultranest_result):
        kwargs = {}
        kwargs["nfev"] = ultranest_result["ncall"]
        kwargs["success"] = ultranest_result["insertion_order_MWW_test"]["converged"]
        kwargs["samples"] = ultranest_result["samples"]
        kwargs["sampler_results"] = ultranest_result
        return cls(**kwargs)


class SamplerLikelihood:
    """Wrapper of the likelihood function used by the sampler.

    This is needed to modify parameters and likelihood by *-0.5

    Parameters
    ----------
    parameters : `~gammapy.modeling.Parameters`
        Parameters with starting values.
    function : callable
        Likelihood function.
    """

    # TODO: Will be updated with the FitStatistic class when ready.

    def __init__(self, function, parameters):
        self.function = function
        self.parameters = parameters

    def fcn(self, value):
        self.parameters.value = value
        total_stat = -0.5 * self.function()
        return total_stat



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Command line tool to check various things about a Gammapy installation.

This file is called `check` instead of `test` to prevent confusion
for developers and the test runner from including it in test collection.
"""
import logging
import warnings
import click

log = logging.getLogger(__name__)


@click.group("check")
def cli_check():
    """Run checks for Gammapy."""


@cli_check.command("logging")
def cli_check_logging():
    """Check logging."""
    log.debug("this is log.debug() output")
    log.info("this is log.info() output")
    log.warning("this is log.warning() output")
    warnings.warn("this is warnings.warn() output")
    print("this is stdout output")



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
from gammapy.utils.array import array_stats_str, shape_2N


def test_array_stats_str():
    actual = array_stats_str(np.pi, "pi")
    assert actual == "pi             : size =     1, min =  3.142, max =  3.142\n"

    actual = array_stats_str([np.pi, 42])
    assert actual == "size =     2, min =  3.142, max = 42.000\n"


def test_shape_2N():
    shape = (34, 89, 120, 444)
    expected_shape = (40, 96, 128, 448)
    assert expected_shape == shape_2N(shape=shape, N=3)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import logging
import numpy as np
from numpy.testing import assert_allclose, assert_equal
import astropy.units as u
from astropy.table import Table
from astropy.time import Time
from astropy.visualization import quantity_support
import matplotlib.pyplot as plt
from gammapy.data import GTI
from gammapy.maps import LabelMapAxis, MapAxes, MapAxis, RegionNDMap, TimeMapAxis
from gammapy.utils.scripts import make_path
from gammapy.utils.testing import assert_time_allclose, mpl_plot_check, requires_data
from gammapy.utils.time import time_ref_to_dict

MAP_AXIS_INTERP = [
    (np.array([0.25, 0.75, 1.0, 2.0]), "lin"),
    (np.array([0.25, 0.75, 1.0, 2.0]), "log"),
    (np.array([0.25, 0.75, 1.0, 2.0]), "sqrt"),
]

MAP_AXIS_NODE_TYPES = [
    ([0.25, 0.75, 1.0, 2.0], "lin", "edges"),
    ([0.25, 0.75, 1.0, 2.0], "log", "edges"),
    ([0.25, 0.75, 1.0, 2.0], "sqrt", "edges"),
    ([0.25, 0.75, 1.0, 2.0], "lin", "center"),
    ([0.25, 0.75, 1.0, 2.0], "log", "center"),
    ([0.25, 0.75, 1.0, 2.0], "sqrt", "center"),
]

nodes_array = np.array([0.25, 0.75, 1.0, 2.0])

MAP_AXIS_NODE_TYPE_UNIT = [
    (nodes_array, "lin", "edges", "s", "TEST", True),
    (nodes_array, "log", "edges", "s", "test", False),
    (nodes_array, "lin", "edges", "TeV", "TEST", False),
    (nodes_array, "sqrt", "edges", "s", "test", False),
    (nodes_array, "lin", "center", "s", "test", False),
    (nodes_array + 1e-9, "lin", "edges", "s", "test", True),
    (nodes_array + 1e-3, "lin", "edges", "s", "test", False),
    (nodes_array / 3600.0, "lin", "edges", "hr", "TEST", True),
]


@pytest.fixture
def time_intervals():
    t0 = Time("2020-03-19")
    t_min = np.linspace(0, 10, 20) * u.d
    t_max = t_min + 1 * u.h
    return {"t_min": t_min, "t_max": t_max, "t_ref": t0}


@pytest.fixture
def time_interval():
    t0 = Time("2020-03-19")
    t_min = 1 * u.d
    t_max = 11 * u.d
    return {"t_min": t_min, "t_max": t_max, "t_ref": t0}


@pytest.fixture(scope="session")
def energy_axis_ref():
    edges = np.arange(1, 11) * u.TeV
    return MapAxis.from_edges(edges, name="energy")


def test_mapaxis_repr():
    axis = MapAxis([1, 2, 3], name="test")
    assert "MapAxis" in repr(axis)


def test_mapaxis_invalid_name():
    with pytest.raises(TypeError):
        MapAxis([1, 2, 3], name=1)


@pytest.mark.parametrize(
    ("nodes", "interp", "node_type", "unit", "name", "result"),
    MAP_AXIS_NODE_TYPE_UNIT,
)
def test_mapaxis_equal(nodes, interp, node_type, unit, name, result):
    axis1 = MapAxis(
        nodes=[0.25, 0.75, 1.0, 2.0],
        name="test",
        unit="s",
        interp="lin",
        node_type="edges",
    )

    axis2 = MapAxis(nodes, name=name, unit=unit, interp=interp, node_type=node_type)

    assert (axis1 == axis2) is result
    assert (axis1 != axis2) is not result


def test_squash():
    axis = MapAxis(
        nodes=[0, 1, 2, 3], unit="TeV", name="energy", node_type="edges", interp="lin"
    )
    ax_sq = axis.squash()

    assert_allclose(ax_sq.nbin, 1)
    assert_allclose(axis.edges[0], ax_sq.edges[0])
    assert_allclose(axis.edges[-1], ax_sq.edges[1])
    assert_allclose(ax_sq.center, 1.5 * u.TeV)


def test_upsample():
    axis = MapAxis(
        nodes=[0, 1, 2, 3], unit="TeV", name="energy", node_type="edges", interp="lin"
    )
    axis_up = axis.upsample(10)

    assert_allclose(axis_up.nbin, 10 * axis.nbin)
    assert_allclose(axis_up.edges[0], axis.edges[0])
    assert_allclose(axis_up.edges[-1], axis.edges[-1])
    assert axis_up.node_type == axis.node_type


def test_downsample():
    axis = MapAxis(
        nodes=[0, 1, 2, 3, 4, 5, 6, 7, 8],
        unit="TeV",
        name="energy",
        node_type="edges",
        interp="lin",
    )

    # Node_type edge, divisible, strict=True
    axis_down = axis.downsample(2)

    assert_allclose(axis_down.nbin, 0.5 * axis.nbin)
    assert_allclose(axis_down.edges[0], axis.edges[0])
    assert_allclose(axis_down.edges[-1], axis.edges[-1])
    assert axis_down.node_type == axis.node_type

    # Node_type edge, divisible, strict=False
    axis_down1 = axis.downsample(2, strict=False)
    assert axis_down == axis_down1

    axis = MapAxis(
        nodes=[0, 1, 2, 3, 4, 5, 6, 7],
        unit="TeV",
        name="energy",
        node_type="edges",
        interp="lin",
    )
    # Node_type edge, not divisible, strict=True
    with pytest.raises(ValueError) as exc_info:
        axis.downsample(2, strict=True)
    assert str(exc_info.value) == "Number of energy bins (7) is not divisible by 2"

    # Node_type edge, not divisible, strict=False
    axis_down = axis.downsample(2, strict=False)
    assert_allclose(axis_down.nbin, np.ceil(0.5 * axis.nbin))
    assert_allclose(axis_down.edges[0], axis.edges[0])
    assert_allclose(axis_down.edges[-1], axis.edges[-1])
    assert axis_down.node_type == axis.node_type

    axis = MapAxis(
        nodes=[
            0,
            1,
            2,
            3,
            4,
            5,
        ],
        unit="TeV",
        name="energy",
        node_type="center",
        interp="lin",
    )

    # Node_type center, not divisible, strict=True
    with pytest.raises(ValueError) as exc_info:
        axis.downsample(2, strict=True)
    assert str(exc_info.value) == "Number of energy bins - 1 (5) is not divisible by 2"
    # Node_type center, not divisible, strict=False
    axis_down = axis.downsample(2, strict=False)
    assert_allclose(axis_down.nbin, 4)
    assert_allclose(axis_down.center, [0.0, 2.0, 4.0, 5] * u.TeV)
    assert axis_down.node_type == "center"

    axis = MapAxis(
        nodes=[
            0,
            1,
            2,
            3,
            4,
        ],
        unit="TeV",
        name="energy",
        node_type="center",
        interp="lin",
    )
    # Node_type center, divisible, strict=True
    axis_down = axis.downsample(2, strict=True)
    assert_allclose(axis_down.nbin, 3)
    assert_allclose(axis_down.center, [0.0, 2.0, 4.0] * u.TeV)
    # Node_type center, divisible, strict=False
    axis_down1 = axis.downsample(2, strict=False)
    assert axis_down == axis_down1


def test_upsample_non_regular():
    axis = MapAxis.from_edges([0, 1, 3, 7], name="test", interp="lin")
    axis_up = axis.upsample(2)

    assert_allclose(axis_up.nbin, 2 * axis.nbin)
    assert_allclose(axis_up.edges[0], axis.edges[0])
    assert_allclose(axis_up.edges[-1], axis.edges[-1])
    assert axis_up.node_type == axis.node_type


def test_upsample_non_regular_nodes():
    axis = MapAxis.from_nodes([0, 1, 3, 7], name="test", interp="lin")
    axis_up = axis.upsample(2)

    assert_allclose(axis_up.nbin, 2 * axis.nbin - 1)
    assert_allclose(axis_up.center[0], axis.center[0])
    assert_allclose(axis_up.center[-1], axis.center[-1])
    assert axis_up.node_type == axis.node_type


def test_downsample_non_regular():
    axis = MapAxis.from_edges([0, 1, 3, 7, 13], name="test", interp="lin")
    axis_down = axis.downsample(2)

    assert_allclose(axis_down.nbin, 0.5 * axis.nbin)
    assert_allclose(axis_down.edges[0], axis.edges[0])
    assert_allclose(axis_down.edges[-1], axis.edges[-1])
    assert axis_down.node_type == axis.node_type


def test_downsample_non_regular_nodes():
    axis = MapAxis.from_edges([0, 1, 3, 7, 9], name="test", interp="lin")
    axis_down = axis.downsample(2)

    assert_allclose(axis_down.nbin, 0.5 * axis.nbin)
    assert_allclose(axis_down.edges[0], axis.edges[0])
    assert_allclose(axis_down.edges[-1], axis.edges[-1])
    assert axis_down.node_type == axis.node_type


@pytest.mark.parametrize("factor", [1, 3, 5, 7, 11])
def test_up_downsample_consistency(factor):
    axis = MapAxis.from_edges([0, 1, 3, 7, 13], name="test", interp="lin")
    axis_new = axis.upsample(factor).downsample(factor)
    assert_allclose(axis.edges, axis_new.edges)


def test_one_bin_nodes():
    axis = MapAxis.from_nodes([1], name="test", unit="deg")

    assert_allclose(axis.center, 1 * u.deg)
    assert_allclose(axis.coord_to_pix(1 * u.deg), 0)
    assert_allclose(axis.coord_to_pix(2 * u.deg), 0)
    assert_allclose(axis.pix_to_coord(0), 1 * u.deg)


def test_table():
    axis = MapAxis(
        nodes=[0, 1, 2, 3, 4, 5, 6, 7, 8],
        unit="TeV",
        name="energy",
        node_type="edges",
        interp="lin",
    )

    table = axis.to_table()
    assert table.colnames == ["CHANNEL", "E_MIN", "E_MAX"]
    assert len(table["CHANNEL"]) == axis.nbin
    assert table["E_MIN"].unit == table["E_MAX"].unit == u.TeV
    assert (table["E_MIN"].data == axis.edges_min.value).all()

    table_kev = axis.to_table("ogip-sherpa")
    assert table_kev["E_MIN"].unit == table_kev["E_MAX"].unit == u.keV


def test_group_table_basic(energy_axis_ref):
    energy_edges = [1, 2, 10] * u.TeV

    groups = energy_axis_ref.group_table(energy_edges)

    assert_allclose(groups["group_idx"], [0, 1])
    assert_allclose(groups["idx_min"], [0, 1])
    assert_allclose(groups["idx_max"], [0, 8])
    assert_allclose(groups["energy_min"], [1, 2])
    assert_allclose(groups["energy_max"], [2, 10])

    bin_type = [_.strip() for _ in groups["bin_type"]]
    assert_equal(bin_type, ["normal", "normal"])


@pytest.mark.parametrize(
    "energy_edges",
    [[1.8, 4.8, 7.2] * u.TeV, [2, 5, 7] * u.TeV, [2000, 5000, 7000] * u.GeV],
)
def test_group_tablenergy_edges(energy_axis_ref, energy_edges):
    groups = energy_axis_ref.group_table(energy_edges)

    assert_allclose(groups["group_idx"], [0, 1, 2, 3])
    assert_allclose(groups["idx_min"], [0, 1, 4, 6])
    assert_allclose(groups["idx_max"], [0, 3, 5, 8])
    assert_allclose(groups["energy_min"].quantity.to_value("TeV"), [1, 2, 5, 7])
    assert_allclose(groups["energy_max"].quantity.to_value("TeV"), [2, 5, 7, 10])

    bin_type = [_.strip() for _ in groups["bin_type"]]
    assert_equal(bin_type, ["underflow", "normal", "normal", "overflow"])


def test_group_table_below_range(energy_axis_ref):
    energy_edges = [0.7, 0.8, 1, 4] * u.TeV
    groups = energy_axis_ref.group_table(energy_edges)

    assert_allclose(groups["group_idx"], [0, 1])
    assert_allclose(groups["idx_min"], [0, 3])
    assert_allclose(groups["idx_max"], [2, 8])
    assert_allclose(groups["energy_min"], [1, 4])
    assert_allclose(groups["energy_max"], [4, 10])

    bin_type = [_.strip() for _ in groups["bin_type"]]
    assert_equal(bin_type, ["normal", "overflow"])


def test_group_table_above_range(energy_axis_ref):
    energy_edges = [5, 7, 11, 13] * u.TeV
    groups = energy_axis_ref.group_table(energy_edges)

    assert_allclose(groups["group_idx"], [0, 1, 2])
    assert_allclose(groups["idx_min"], [0, 4, 6])
    assert_allclose(groups["idx_max"], [3, 5, 8])
    assert_allclose(groups["energy_min"], [1, 5, 7])
    assert_allclose(groups["energy_max"], [5, 7, 10])

    bin_type = [_.strip() for _ in groups["bin_type"]]
    assert_equal(bin_type, ["underflow", "normal", "normal"])


def test_group_table_outside_range(energy_axis_ref):
    energy_edges = [20, 30, 40] * u.TeV

    with pytest.raises(ValueError):
        energy_axis_ref.group_table(energy_edges)


def test_map_axis_aligned():
    ax1 = MapAxis([1, 2, 3], interp="lin", node_type="edges")
    ax2 = MapAxis([1.5, 2.5], interp="log", node_type="center")
    assert not ax1.is_aligned(ax2)


def test_map_axis_pad():
    axis = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=1)

    padded = axis.pad(pad_width=(0, 1))
    assert_allclose(padded.edges, [1, 10, 100] * u.TeV)

    padded = axis.pad(pad_width=(1, 0))
    assert_allclose(padded.edges, [0.1, 1, 10] * u.TeV)

    padded = axis.pad(pad_width=1)
    assert_allclose(padded.edges, [0.1, 1, 10, 100] * u.TeV)


def test_map_axes_pad():
    axis_1 = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=1)
    axis_2 = MapAxis.from_bounds(0, 1, nbin=2, unit="deg", name="rad")

    axes = MapAxes([axis_1, axis_2])

    axes = axes.pad(axis_name="energy", pad_width=1)

    assert_allclose(axes["energy"].edges, [0.1, 1, 10, 100] * u.TeV)


def test_rename():
    axis_1 = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=1)
    axis = axis_1.rename("energy_true")
    assert axis_1.name == "energy"
    assert axis.name == "energy_true"

    axis_2 = MapAxis.from_bounds(0, 1, nbin=2, unit="deg", name="rad")

    axes = MapAxes([axis_1, axis_2])
    axes = axes.rename_axes(["energy", "rad"], ["energy_true", "angle"])
    assert axes.names == ["energy_true", "angle"]


@pytest.mark.parametrize(("edges", "interp"), MAP_AXIS_INTERP)
def test_mapaxis_init_from_edges(edges, interp):
    axis = MapAxis(edges, interp=interp)
    assert_allclose(axis.edges, edges)
    assert_allclose(axis.nbin, len(edges) - 1)
    with pytest.raises(ValueError):
        MapAxis.from_edges([1])
        MapAxis.from_edges([0, 1, 1, 2])
        MapAxis.from_edges([0, 1, 3, 2])


@pytest.mark.parametrize(("nodes", "interp"), MAP_AXIS_INTERP)
def test_mapaxis_from_nodes(nodes, interp):
    axis = MapAxis.from_nodes(nodes, interp=interp)
    assert_allclose(axis.center, nodes)
    assert_allclose(axis.nbin, len(nodes))
    with pytest.raises(ValueError):
        MapAxis.from_nodes([])
        MapAxis.from_nodes([0, 1, 1, 2])
        MapAxis.from_nodes([0, 1, 3, 2])


@pytest.mark.parametrize(("nodes", "interp"), MAP_AXIS_INTERP)
def test_mapaxis_from_bounds(nodes, interp):
    axis = MapAxis.from_bounds(nodes[0], nodes[-1], 3, interp=interp)
    assert_allclose(axis.edges[0], nodes[0])
    assert_allclose(axis.edges[-1], nodes[-1])
    assert_allclose(axis.nbin, 3)
    with pytest.raises(ValueError):
        MapAxis.from_bounds(1, 1, 1)


def test_map_axis_from_energy_units():
    with pytest.raises(ValueError):
        _ = MapAxis.from_energy_bounds(0.1, 10, 2, unit="deg")

    with pytest.raises(ValueError):
        _ = MapAxis.from_energy_edges([0.1, 1, 10] * u.deg)


@pytest.mark.parametrize(("nodes", "interp", "node_type"), MAP_AXIS_NODE_TYPES)
def test_mapaxis_pix_to_coord(nodes, interp, node_type):
    axis = MapAxis(nodes, interp=interp, node_type=node_type)
    assert_allclose(axis.center, axis.pix_to_coord(np.arange(axis.nbin, dtype=float)))
    assert_allclose(
        np.arange(axis.nbin + 1, dtype=float) - 0.5, axis.coord_to_pix(axis.edges)
    )


@pytest.mark.parametrize(("nodes", "interp", "node_type"), MAP_AXIS_NODE_TYPES)
def test_mapaxis_coord_to_idx(nodes, interp, node_type):
    axis = MapAxis(nodes, interp=interp, node_type=node_type)
    assert_allclose(np.arange(axis.nbin, dtype=int), axis.coord_to_idx(axis.center))


@pytest.mark.parametrize(("nodes", "interp", "node_type"), MAP_AXIS_NODE_TYPES)
def test_mapaxis_slice(nodes, interp, node_type):
    axis = MapAxis(nodes, interp=interp, node_type=node_type)
    saxis = axis.slice(slice(1, 3))
    assert_allclose(saxis.nbin, 2)
    assert_allclose(saxis.center, axis.center[slice(1, 3)])

    axis = MapAxis(nodes, interp=interp, node_type=node_type)
    saxis = axis.slice(slice(1, None))
    assert_allclose(saxis.nbin, axis.nbin - 1)
    assert_allclose(saxis.center, axis.center[slice(1, None)])

    axis = MapAxis(nodes, interp=interp, node_type=node_type)
    saxis = axis.slice(slice(None, 2))
    assert_allclose(saxis.nbin, 2)
    assert_allclose(saxis.center, axis.center[slice(None, 2)])

    axis = MapAxis(nodes, interp=interp, node_type=node_type)
    saxis = axis.slice(slice(None, -1))
    assert_allclose(saxis.nbin, axis.nbin - 1)
    assert_allclose(saxis.center, axis.center[slice(None, -1)])


def test_map_axis_plot_helpers():
    axis = MapAxis.from_nodes([0, 1, 2], unit="deg", name="offset")
    labels = axis.as_plot_labels

    assert labels[0] == "0.00e+00 deg"

    assert_allclose(axis.center, axis.as_plot_center)
    assert_allclose(axis.edges, axis.as_plot_edges)


def test_map_axis_concatenate():
    axis_1 = MapAxis.from_bounds(0, 10, 10, name="axis")
    axis_2 = MapAxis.from_bounds(10, 20, 10, name="axis")
    axis_2_other_name = MapAxis.from_bounds(10, 20, 10, name="other_axis")

    axis_12 = axis_1.concatenate(axis_2)

    assert_equal(axis_12.edges, np.linspace(0, 20, 21))

    with pytest.raises(ValueError):
        axis_1.concatenate(axis_2_other_name)


def test_time_axis(time_intervals):
    axis = TimeMapAxis(
        time_intervals["t_min"], time_intervals["t_max"], time_intervals["t_ref"]
    )

    axis_copy = axis.copy()

    assert axis.nbin == 20
    assert axis.name == "time"
    assert axis.node_type == "intervals"

    assert_allclose(axis.time_delta.to_value("min"), 60)
    assert_allclose(axis.time_mid[0].mjd, 58927.020833333336)

    assert "time" in axis.__str__()
    assert "20" in axis.__str__()

    with pytest.raises(ValueError):
        axis.assert_name("bad")

    assert axis_copy == axis

    assert not axis.is_contiguous

    ax_cont = axis.to_contiguous()
    assert_allclose(ax_cont.nbin, 39)


def test_single_interval_time_axis(time_interval):
    axis = TimeMapAxis(
        edges_min=time_interval["t_min"],
        edges_max=time_interval["t_max"],
        reference_time=time_interval["t_ref"],
    )

    coord = Time(58933, format="mjd") + u.Quantity([1.5, 3.5, 10], unit="d")
    pix = axis.coord_to_pix(coord)

    assert axis.nbin == 1
    assert_allclose(axis.time_delta.to_value("d"), 10)
    assert_allclose(axis.time_mid[0].mjd, 58933)

    pix_min = axis.coord_to_pix(time_interval["t_min"] + 0.001 * u.s)
    assert_allclose(pix_min, -0.5)

    pix_max = axis.coord_to_pix(time_interval["t_max"] - 0.001 * u.s)
    assert_allclose(pix_max, 0.5)

    assert_allclose(pix, [0.15, 0.35, np.nan])


def test_slice_squash_time_axis(time_intervals, caplog):
    axis = TimeMapAxis(
        time_intervals["t_min"], time_intervals["t_max"], time_intervals["t_ref"]
    )
    axis_squash = axis.squash()
    axis_slice = axis.slice(slice(1, 5))

    assert axis_squash.nbin == 1
    assert_allclose(axis_squash.time_min[0].mjd, 58927)
    assert_allclose(axis_squash.time_delta.to_value("d"), 10.04166666)
    assert axis_slice.nbin == 4
    assert_allclose(axis_slice.time_delta.to_value("d")[0], 0.04166666666)
    caplog.set_level(logging.DEBUG)

    assert axis_squash != axis_slice
    assert caplog.records[0].message == "TimeMapAxis time is not equal"
    assert caplog.records[0].levelname == "DEBUG"


def test_from_time_edges_time_axis():
    t0 = Time("2020-03-19")
    t_min = t0 + np.linspace(0, 10, 20) * u.d
    t_max = t_min + 1 * u.h

    axis = TimeMapAxis.from_time_edges(t_min, t_max)
    axis_h = TimeMapAxis.from_time_edges(t_min, t_max, unit="h")

    assert axis.nbin == 20
    assert axis.name == "time"
    assert_time_allclose(axis.reference_time, t0)
    assert_allclose(axis.time_delta.to_value("min"), 60)
    assert_allclose(axis.time_mid[0].mjd, 58927.020833333336)
    assert_allclose(axis_h.time_delta.to_value("h"), 1)
    assert_allclose(axis_h.time_mid[0].mjd, 58927.020833333336)
    assert axis == axis_h


def test_incorrect_time_axis():
    tmin = np.linspace(0, 10, 20) * u.h
    tmax = np.linspace(1, 11, 20) * u.h

    # incorrect reference time
    with pytest.raises(ValueError):
        TimeMapAxis(tmin, tmax, reference_time=51000 * u.d, name="time")

    # overlapping time intervals
    with pytest.raises(ValueError):
        TimeMapAxis(tmin, tmax, reference_time=Time.now(), name="time")


def test_bad_length_sort_time_axis(time_intervals):
    tref = time_intervals["t_ref"]
    tmin = time_intervals["t_min"]
    tmax_reverse = time_intervals["t_max"][::-1]
    tmax_short = time_intervals["t_max"][:-1]

    with pytest.raises(ValueError):
        TimeMapAxis(tmin, tmax_reverse, tref, name="time")

    with pytest.raises(ValueError):
        TimeMapAxis(tmin, tmax_short, tref, name="time")


def test_coord_to_idx_time_axis(time_intervals):
    tmin = time_intervals["t_min"]
    tmax = time_intervals["t_max"]
    tref = time_intervals["t_ref"]
    axis = TimeMapAxis(tmin, tmax, tref, name="time")

    time = Time(58927.020833333336, format="mjd")

    times = axis.time_mid
    times[::2] += 1 * u.h
    times = times.insert(0, tref - [1, 2] * u.yr)

    idx = axis.coord_to_idx(time)
    indices = axis.coord_to_idx(times)

    pix = axis.coord_to_pix(time)
    pixels = axis.coord_to_pix(times)

    assert idx == 0
    assert_allclose(indices[1::2], [-1, 1, 3, 5, 7, 9, 11, 13, 15, 17, 19])
    assert_allclose(indices[::2], -1)
    assert_allclose(pix, 0, atol=1e-10)
    assert_allclose(pixels[1::2], [np.nan, 1, 3, 5, 7, 9, 11, 13, 15, 17, 19])


def test_timeaxis_table(time_intervals):
    tmin = time_intervals["t_min"]
    tmax = time_intervals["t_max"]
    tref = time_intervals["t_ref"]
    axis = TimeMapAxis(tmin, tmax, tref, name="time")

    table = axis.to_table()

    assert table.colnames == ["START", "STOP"]
    assert len(table["START"]) == axis.nbin
    assert (table["START"] == axis.time_min).all()


def test_pix_to_coord_time_axis(time_intervals):
    tmin = time_intervals["t_min"]
    tmax = time_intervals["t_max"]
    tref = time_intervals["t_ref"]
    axis = TimeMapAxis(tmin, tmax, tref, name="time")

    pixels = [1.3, 3.2, 5.4, 7, 15.33, 17.21, 19.11]
    coords = axis.pix_to_coord(pixels)
    assert_allclose(
        coords[0:3].mjd, [58927.538816, 58928.587281, 58929.648246], rtol=1e-5
    )

    # test with nan indices
    pixels.append(np.nan)
    coords = axis.pix_to_coord(pixels)
    assert_allclose(
        coords[-3:].mjd,
        [58935.9561, 58937.0046, -3.72500000e-04],
        rtol=1e-5,
    )

    # assert with invalid pixels & multidim array
    coords = axis.pix_to_coord([[-1.2, 0.6], [1.5, 24.7]])
    assert_allclose(
        coords.mjd,
        [[-3.72500000e-04, 5.89270250e04], [5.89275471e04, -3.72500000e-04]],
        rtol=1e-5,
    )

    # test with one value
    coords = axis.pix_to_coord(3)
    assert_allclose(coords.mjd, axis.time_min[3].mjd, rtol=1e-5)


def test_slice_time_axis(time_intervals):
    axis = TimeMapAxis(
        time_intervals["t_min"], time_intervals["t_max"], time_intervals["t_ref"]
    )

    new_axis = axis.slice([2, 6, 9])
    squashed = axis.squash()

    assert new_axis.nbin == 3
    assert_allclose(squashed.time_max[0].mjd, 58937.041667)
    assert squashed.nbin == 1
    assert_allclose(squashed.time_max[0].mjd, 58937.041667)


def test_time_map_axis_from_time_bounds():
    t_min = Time("2006-02-12", scale="utc")
    t_max = t_min + 12 * u.h

    axis = TimeMapAxis.from_time_bounds(time_min=t_min, time_max=t_max, nbin=3)
    assert_allclose(axis.center, [0.083333, 0.25, 0.416667] * u.d, rtol=1e-5)


def test_from_table_time_axis():
    t0 = Time("2006-02-12", scale="utc")
    t_min = np.linspace(0, 10, 10) * u.d
    t_max = t_min + 12 * u.h

    table = Table()
    table["TIME_MIN"] = t_min
    table["TIME_MAX"] = t_max
    table.meta.update(time_ref_to_dict(t0))
    table.meta["AXCOLS1"] = "TIME_MIN,TIME_MAX"

    axis = TimeMapAxis.from_table(table, format="gadf")

    assert axis.nbin == 10
    assert_allclose(axis.time_mid[0].mjd, 53778.25)


def test_from_table_time_axis_lightcurve_format():
    t0 = Time("2006-02-12", scale="tt")
    t_min = np.linspace(0, 10, 10) * u.d
    t_max = t_min + 12 * u.h

    table = Table()
    table["time_min"] = t_min.to_value("h")
    table["time_max"] = t_max.to_value("h")
    table.meta.update(time_ref_to_dict(t0))
    table.meta["TIMEUNIT"] = "h"

    axis = TimeMapAxis.from_table(table, format="lightcurve")

    assert axis.nbin == 10
    assert_allclose(axis.time_mid[0].mjd, 53778.25)
    assert axis.time_mid.scale == "tt"
    t0.format = "mjd"
    assert_time_allclose(axis.reference_time, t0)


@requires_data()
def test_from_gti_time_axis():
    filename = "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz"
    filename = make_path(filename)
    gti = GTI.read(filename)

    axis = TimeMapAxis.from_gti(gti)
    expected = Time(53090.123451203704, format="mjd", scale="tt")
    assert_time_allclose(axis.time_min[0], expected)
    assert axis.nbin == 1


def test_from_gti_bounds():
    start = u.Quantity([1, 2], "min")
    stop = u.Quantity([1.5, 2.5], "min")
    time_ref = Time("2010-01-01 00:00:00.0")

    gti = GTI.create(start, stop, time_ref)

    axis = TimeMapAxis.from_gti_bounds(
        gti=gti,
        t_delta=10 * u.s,
    )

    assert axis.nbin == 8
    expected = Time("2010-01-01 00:01:00.0")
    # GTI.create() changes the reference time format
    expected.format = "mjd"

    assert_time_allclose(axis.time_min[0], expected)

    expected = Time("2010-01-01 00:02:30.0")
    expected.format = "mjd"
    assert_time_allclose(axis.time_max[-1], expected)


def test_to_gti(time_intervals):
    time_axis = TimeMapAxis(
        time_intervals["t_min"], time_intervals["t_max"], time_intervals["t_ref"]
    )
    gti = time_axis.to_gti()
    assert len(gti.table) == 20
    assert TimeMapAxis.from_gti(gti) == time_axis


def test_map_with_time_axis(time_intervals):
    time_axis = TimeMapAxis(
        time_intervals["t_min"], time_intervals["t_max"], time_intervals["t_ref"]
    )
    energy_axis = MapAxis.from_energy_bounds(0.1, 10, 2, unit="TeV")
    region_map = RegionNDMap.create(
        region="fk5; circle(0,0,0.1)", axes=[energy_axis, time_axis]
    )

    assert region_map.geom.data_shape == (20, 2, 1, 1)


def test_time_axis_plot_helpers():
    time_ref = Time("1999-01-01T00:00:00.123456789")

    time_axis = TimeMapAxis(
        edges_min=[0, 1, 3] * u.d,
        edges_max=[0.8, 1.9, 5.4] * u.d,
        reference_time=time_ref,
    )

    labels = time_axis.as_plot_labels
    assert labels[0] == "1999-01-01 00:00:00.123 - 1999-01-01 19:12:00.123"

    center = time_axis.as_plot_center
    assert center[0].year == 1999

    edges = time_axis.to_contiguous().as_plot_edges
    assert edges[0].year == 1999


def test_axes_basics(caplog):
    energy_axis = MapAxis.from_energy_edges([1, 3] * u.TeV)

    time_ref = Time("1999-01-01T00:00:00.123456789")

    time_axis = TimeMapAxis(
        edges_min=[0, 1, 3] * u.d,
        edges_max=[0.8, 1.9, 5.4] * u.d,
        reference_time=time_ref,
    )

    axes = MapAxes([energy_axis, time_axis])

    assert axes.shape == (1, 3)
    assert axes.is_unidimensional
    assert not axes.is_flat

    assert axes.primary_axis.name == "time"

    new_axes = axes.copy()
    assert new_axes[0] == new_axes[0]
    assert new_axes[1] == new_axes[1]
    assert new_axes == axes

    energy_axis = MapAxis.from_energy_edges([1, 4] * u.TeV)
    new_axes = MapAxes([energy_axis, time_axis.copy()])
    caplog.set_level(logging.DEBUG)
    assert new_axes != axes
    assert "MapAxis energy is not equal" in [_.message for _ in caplog.records]
    assert "DEBUG" in [_.levelname for _ in caplog.records]


def test_map_axes_assert_names():
    axis_a = MapAxis([1, 2, 3], name="axis_a")
    axis_b = MapAxis([1, 2, 3, 4], name="axis_b")
    axis_c = LabelMapAxis(["a", "b"], name="axis_c")

    axes = MapAxes([axis_a, axis_b, axis_c])

    axes.assert_names(["axis_a", "axis_b", "axis_c"])
    axes.assert_names(["axis_a", "axis_b"], allow_extra=True)

    with pytest.raises(ValueError):
        axes.assert_names(["axis_a", "axis_b"])

    with pytest.raises(ValueError):
        axes.assert_names(["axis_c", "axis_b", "axis_a"])

    with pytest.raises(ValueError):
        axes.assert_names(["axis_b"], allow_extra=True)


def test_axes_getitem():
    axis1 = MapAxis.from_bounds(1, 4, 3, name="a1")
    axis2 = axis1.copy(name="a2")
    axis3 = axis1.copy(name="a3")
    axes = MapAxes([axis1, axis2, axis3])

    assert isinstance(axes[0], MapAxis)
    assert axes[-1].name == "a3"
    assert isinstance(axes[1:], MapAxes)
    assert len(axes[1:]) == 2
    assert isinstance(axes[0:1], MapAxes)
    assert len(axes[0:1]) == 1
    assert isinstance(axes[["a3", "a1"]], MapAxes)
    assert axes[["a3", "a1"]][0].name == "a3"


def test_label_map_axis_basics():
    axis = LabelMapAxis(labels=["label-1", "label-2"], name="label-axis")

    axis_str = str(axis)
    assert "node type" in axis_str
    assert "labels" in axis_str
    assert "label-2" in axis_str

    with pytest.raises(ValueError):
        axis.assert_name("time")

    assert axis.nbin == 2
    assert axis.node_type == "label"

    assert_allclose(axis.bin_width, 1)

    assert axis.name == "label-axis"

    with pytest.raises(ValueError):
        axis.edges

    axis_copy = axis.copy()
    assert axis_copy.name == "label-axis"

    # Ensure order is correct
    labels = np.array([1, 4, 5])
    assert_allclose(labels, [1, 4, 5])
    assert_allclose(LabelMapAxis(labels).center, labels)
    assert_allclose(LabelMapAxis(labels)._labels, labels)

    # Test case with non unique labels
    with pytest.raises(ValueError) as exc_info:
        LabelMapAxis([1, 1, 1])
    assert str(exc_info.value) == "Node labels must be unique"

    # Ensure non alphabetical ordering
    labels = np.array(["b", "a", "d", "c"])
    assert_equal(LabelMapAxis(labels).center, labels)


def test_label_map_axis_coord_to_idx():
    axis = LabelMapAxis(labels=["label-1", "label-2", "label-3"], name="label-axis")

    labels = "label-1"
    idx = axis.coord_to_idx(coord=labels)
    assert_allclose(idx, 0)

    labels = ["label-1", "label-3"]
    idx = axis.coord_to_idx(coord=labels)
    assert_allclose(idx, [0, 2])

    labels = [["label-1"], ["label-2"]]
    idx = axis.coord_to_idx(coord=labels)
    assert_allclose(idx, [[0], [1]])

    with pytest.raises(ValueError):
        labels = [["bad-label"], ["label-2"]]
        _ = axis.coord_to_idx(coord=labels)


def test_mixed_axes():
    label_axis = LabelMapAxis(labels=["label-1", "label-2", "label-3"], name="label")

    time_axis = TimeMapAxis(
        edges_min=[1, 10] * u.day,
        edges_max=[2, 13] * u.day,
        reference_time=Time("2020-03-19"),
    )

    energy_axis = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=4)

    axes = MapAxes(axes=[energy_axis, time_axis, label_axis])

    coords = axes.get_coord()

    assert coords["label"].shape == (1, 1, 3)
    assert coords["energy"].shape == (4, 1, 1)
    assert coords["time"].shape == (1, 2, 1)

    idx = axes.coord_to_idx(coords)

    assert_allclose(idx[0], np.arange(4).reshape((4, 1, 1)))
    assert_allclose(idx[1], np.arange(2).reshape((1, 2, 1)))
    assert_allclose(idx[2], np.arange(3).reshape((1, 1, 3)))

    hdu = axes.to_table_hdu(format="gadf")

    table = Table.read(hdu)

    assert table["LABEL"].dtype == np.dtype("U7")
    assert len(table) == 24


def test_map_axis_format_plot_xaxis():
    axis = MapAxis.from_energy_bounds(
        "0.03 TeV", "300 TeV", nbin=20, per_decade=True, name="energy_true"
    )

    with mpl_plot_check():
        ax = plt.gca()
        with quantity_support():
            ax.plot(axis.center, np.ones_like(axis.center))

    ax1 = axis.format_plot_xaxis(ax=ax)
    assert ax1.xaxis.units == u.Unit("TeV")
    assert " ".join(ax1.axes.axes.get_xlabel().split()[:2]) == "True Energy"


def test_time_format(time_intervals):
    axis = TimeMapAxis(
        time_intervals["t_min"],
        time_intervals["t_max"],
        time_intervals["t_ref"],
        name="time",
    )
    with pytest.raises(ValueError):
        axis.time_format = "null"


def test_time_map_axis_format_plot_xaxis(time_intervals):
    axis = TimeMapAxis(
        time_intervals["t_min"],
        time_intervals["t_max"],
        time_intervals["t_ref"],
        name="time",
    )

    with mpl_plot_check():
        ax = plt.gca()
        with quantity_support():
            ax.plot(axis.as_plot_center, np.ones_like(axis.center))

    ax1 = axis.format_plot_xaxis(ax=ax)
    assert ax1.axes.axes.get_xlabel().split()[0] == "Time"
    assert ax1.axes.axes.get_xlabel().split()[1] == "[iso]"

    axis.time_format = "mjd"
    with mpl_plot_check():
        ax = plt.gca()
        with quantity_support():
            ax.plot(axis.as_plot_center, np.ones_like(axis.center))
    ax2 = axis.format_plot_xaxis(ax=ax)
    assert ax2.axes.axes.get_xlabel().split()[1] == "[mjd]"


def test_time_group_table(time_intervals):
    axis = TimeMapAxis(
        time_intervals["t_min"],
        time_intervals["t_max"],
        time_intervals["t_ref"],
        name="time",
    )

    groups = axis.group_table(interval_edges=[5 * u.d - 1 * u.h, 6 * u.d + 2 * u.h])
    assert_allclose(groups["idx_min"], [10])
    assert_allclose(groups["idx_max"], [11])
    assert_allclose(groups["time_min"], [58932.263158])
    assert_allclose(groups["time_max"], [58932.83114])

    groups_overflow = axis.group_table(interval_edges=[11 * u.d, 12 * u.d])
    assert_allclose(groups_overflow["idx_min"], [-1])

    groups_timeedges = axis.group_table(
        interval_edges=[
            Time("2020-03-19") + 5 * u.d - 1 * u.h,
            Time("2020-03-19") + 6 * u.d + 2 * u.h,
        ]
    )
    assert_allclose(groups_timeedges["time_min"], [58932.263158])
    assert_allclose(groups_timeedges["time_max"], [58932.83114])

    groups_exactedges = axis.group_table(interval_edges=[3 * u.d, 7 * u.d + 1 * u.h])
    assert_allclose(groups_exactedges["idx_min"], [6])
    assert_allclose(groups_exactedges["idx_max"], [13])


def test_single_valued_axis():
    # this will be interpreted as a scalar value
    # that is against the specifications, but we allow it nevertheless
    theta_values = np.array([0.5]) * u.deg
    table = Table(data=[theta_values, theta_values], names=["THETA_LO", "THETA_HI"])
    _ = MapAxis.from_table(table, format="gadf-dl3", column_prefix="THETA")

    # this is a proper array-like axis with just a single value
    theta_values = np.array([[0.5]]) * u.deg
    table = Table(data=[theta_values, theta_values], names=["THETA_LO", "THETA_HI"])
    _ = MapAxis.from_table(table, format="gadf-dl3", column_prefix="THETA")


def test_label_map_axis_concatenate():
    label1 = LabelMapAxis(["aa", "bb"], name="letters")
    label2 = LabelMapAxis(["cc", "dd"], name="letters")
    label3 = LabelMapAxis(["ee", "ff"], name="other_letters")

    label_append12 = label1.concatenate(label2)

    assert_equal(label_append12.center, np.array(["aa", "bb", "cc", "dd"], dtype="<U2"))
    assert label_append12.name == "letters"
    with pytest.raises(ValueError):
        label2.concatenate(label3)


def test_label_map_axis_from_stack():
    label1 = LabelMapAxis(["a", "b", "c"], name="letters")
    label2 = LabelMapAxis(["d", "e"], name="letters")
    label3 = LabelMapAxis(["f"], name="letters")

    label_stack = LabelMapAxis.from_stack([label1, label2, label3])

    assert_equal(label_stack.center, np.array(["a", "b", "c", "d", "e", "f"]))
    assert label_stack.name == "letters"


def test_label_map_axis_squash():
    label = LabelMapAxis(["a", "b", "c"], name="Letters")
    squash_label = label.squash()

    assert squash_label.nbin == 1
    assert_equal(squash_label.center, np.array(["a...c"]))


def test_energy_bin_per_decade_not_strict_bounds():
    nbin = 5
    axis = MapAxis.from_energy_bounds(
        "0.03 TeV", "333 TeV", nbin=nbin, per_decade=True, strict_bounds=False
    )

    assert_allclose(axis.edges[0:-nbin] * 10.0, axis.edges[nbin:], rtol=1e-5, atol=0)


def test_periodic_map_axis():
    axis1 = MapAxis.from_bounds(-0.5, 0.5, 5, boundary_type="periodic")
    coords = np.array([-1.0, -0.7, 0.1, 1.2])
    assert_allclose(axis1.wrap_coord(coords), [0.0, 0.3, 0.1, 0.2], rtol=1e-5)

    idx = axis1.coord_to_idx(coord=coords)
    assert_allclose(idx, [2, 4, 2, 3], rtol=1e-5)

    pix = axis1.coord_to_pix(coord=coords)
    assert_allclose(pix, [2.0, 3.5, 2.5, 3.0], rtol=1e-5)

    with pytest.raises(ValueError):
        axis1 = MapAxis.from_bounds(-0.5, 0.5, 5, boundary_type="other")
    with pytest.raises(ValueError):
        axis1 = MapAxis.from_bounds(
            -0.5, 0.5, 5, boundary_type="periodic", interp="log"
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
from numpy.testing import assert_allclose
from gammapy.modeling import Parameter, Parameters
from gammapy.modeling.sherpa import optimize_sherpa

pytest.importorskip("sherpa")


class MyDataset:
    def __init__(self, parameters):
        self.parameters = parameters

    def fcn(self):
        x, y, z = [p.value for p in self.parameters]
        x_opt, y_opt, z_opt = 2, 3e5, 4e-5
        x_err, y_err, z_err = 0.2, 3e4, 4e-6
        return (
            ((x - x_opt) / x_err) ** 2
            + ((y - y_opt) / y_err) ** 2
            + ((z - z_opt) / z_err) ** 2
        )


@pytest.fixture()
def pars():
    x = Parameter("x", 2.1)
    y = Parameter("y", 3.1e5, scale=1e5)
    z = Parameter("z", 4.1e-5, scale=1e-5)
    return Parameters([x, y, z])


# TODO: levmar doesn't work yet; needs array of statval as return in likelihood
# method='gridsearch' would require very low tolerance asserts, not added for now


@pytest.mark.parametrize("method", ["moncar", "simplex"])
def test_sherpa(method, pars):
    ds = MyDataset(pars)
    factors, info, _ = optimize_sherpa(function=ds.fcn, parameters=pars, method=method)

    assert info["success"]
    assert_allclose(ds.fcn(), 0, atol=1e-12)

    # Check the result in parameters is OK
    assert_allclose(pars["x"].value, 2)
    assert_allclose(pars["y"].value, 3e5)
    assert_allclose(pars["z"].value, 4e-5)

    # Check that sherpa sees the parameter factors correctly
    assert_allclose(factors, [2, 3, 4])


def test_sherpa_frozen(pars):
    ds = MyDataset(pars)
    pars["y"].frozen = True

    factors, info, _ = optimize_sherpa(function=ds.fcn, parameters=pars)

    assert info["success"]
    assert_allclose(pars["x"].value, 2)
    assert_allclose(pars["y"].value, 3.1e5)
    assert_allclose(pars["z"].value, 4.0e-5)
    assert_allclose(ds.fcn(), 0.11111111, rtol=1e-6)


@pytest.mark.xfail
def test_sherpa_limits(pars):
    ds = MyDataset(pars)
    pars["y"].min = 301000

    factors, info, _ = optimize_sherpa(function=ds.fcn, parameters=pars)

    assert info["success"]

    # Check the result in parameters is OK
    assert_allclose(pars["x"].value, 2, rtol=1e-2)
    assert_allclose(pars["y"].value, 301000, rtol=1e-3)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
import numpy as np
from gammapy.datasets import Datasets
from gammapy.datasets.actors import DatasetsActor
from gammapy.estimators.parameter import ParameterEstimator
from gammapy.estimators.utils import _get_default_norm
from gammapy.maps import Map, MapAxis
from gammapy.modeling.models import ScaleSpectralModel

log = logging.getLogger(__name__)


class FluxEstimator(ParameterEstimator):
    """Flux estimator.

    Estimates flux for a given list of datasets with their model in a given energy range.

    To estimate the model flux the amplitude of the reference spectral model is
    fitted within the energy range. The amplitude is re-normalized using the "norm" parameter,
    which specifies the deviation of the flux from the reference model in this
    energy range.

    Note that there should be only one free norm or amplitude parameter for the estimator to run.

    Parameters
    ----------
    source : str or int
        For which source in the model to compute the flux.
    n_sigma : float, optional
        Sigma to use for asymmetric error computation. Must be a positive value.
        Default is 1.
    n_sigma_ul : float, optional
        Sigma to use for upper limit computation. Must be a positive value.
        Default is 2.
    n_sigma_sensitivity : float, optional
        Sigma to use for sensitivity computation. Must be a positive value.
        Default is 5.
    selection_optional : list of str, optional
        Which additional quantities to estimate. Available options are:

            * "all": all the optional steps are executed.
            * "errn-errp": estimate asymmetric errors.
            * "ul": estimate upper limits.
            * "scan": estimate fit statistic profiles.

        Default is None so the optional steps are not executed.
    fit : `~gammapy.modeling.Fit`, optional
        Fit instance specifying the backend and fit options.
        If None, the `~gammapy.modeling.Fit` instance is created internally. Default is None.
    reoptimize : bool, optional
        If True, the free parameters of the other models are fitted in each bin independently,
        together with the norm of the source of interest
        (but the other parameters of the source of interest are kept frozen).
        If False, only the norm of the source of interest is fitted,
        and all other parameters are frozen at their current values.
        Default is False.
    norm : `~gammapy.modeling.Parameter` or dict, optional
        Norm parameter used for the fit.
        Default is None and a new parameter is created automatically,
        with value=1, name="norm", scan_min=0.2, scan_max=5, and scan_n_values = 11.
        By default, the min and max are not set and derived from the source model,
        unless the source model does not have one and only one norm parameter.
        If a dict is given the entries should be a subset of
        `~gammapy.modeling.Parameter` arguments.
    """

    tag = "FluxEstimator"

    def __init__(
        self,
        source=0,
        n_sigma=1,
        n_sigma_ul=2,
        n_sigma_sensitivity=5,
        selection_optional=None,
        fit=None,
        reoptimize=False,
        norm=None,
    ):
        self.source = source

        scan_n_sigma = np.maximum(n_sigma_ul, n_sigma_sensitivity) + 1
        self.norm = _get_default_norm(norm, scan_n_sigma=scan_n_sigma, interp="log")

        super().__init__(
            null_value=0,
            n_sigma=n_sigma,
            n_sigma_ul=n_sigma_ul,
            n_sigma_sensitivity=n_sigma_sensitivity,
            selection_optional=selection_optional,
            fit=fit,
            reoptimize=reoptimize,
        )

    def get_scale_model(self, models):
        """Set scale model.

        Parameters
        ----------
        models : `~gammapy.modeling.Models`
            Models.

        Returns
        -------
        model : `~gammapy.modeling.ScaleSpectralModel`
            Scale spectral model.
        """
        ref_model = models[self.source].spectral_model
        scale_model = ScaleSpectralModel(ref_model)
        scale_model.norm = self.norm.copy()
        return scale_model

    def estimate_npred_excess(self, datasets):
        """Estimate npred excess for the source.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            Datasets.

        Returns
        -------
        result : dict
            Dictionary with an array with one entry per dataset with the sum of the
            masked npred excess.
        """
        npred_excess = []

        for dataset in datasets:
            name = datasets.models[self.source].name
            npred_signal = dataset.npred_signal(model_names=[name])
            npred = Map.from_geom(dataset.counts.geom)
            npred.stack(npred_signal)
            npred_excess.append(npred.data[dataset.mask].sum())

        return {"npred_excess": np.array(npred_excess), "datasets": datasets.names}

    def run(self, datasets):
        """Estimate flux for a given energy range.

        Parameters
        ----------
        datasets : list of `~gammapy.datasets.SpectrumDataset`
            Spectrum datasets.

        Returns
        -------
        result : dict
            Dictionary with results for the flux point.
        """
        if not isinstance(datasets, DatasetsActor):
            datasets = Datasets(datasets)
        models = datasets.models.copy()

        model = self.get_scale_model(models)

        energy_min, energy_max = datasets.energy_ranges
        energy_axis = MapAxis.from_energy_edges([energy_min.min(), energy_max.max()])
        if np.any(model(energy_axis.edges).value < 0.0):
            log.warning(
                "Reference source model predicts negative flux. Results of estimator should be interpreted with caution"
            )

        with np.errstate(invalid="ignore", divide="ignore"):
            result = model.reference_fluxes(energy_axis=energy_axis)
            # convert to scalar values
            result = {key: value.item() for key, value in result.items()}

        # freeze all source model parameters
        models[self.source].parameters.freeze_all()

        models[self.source].spectral_model = model
        datasets.models = models
        result.update(super().run(datasets, model.norm))

        datasets.models[self.source].spectral_model.norm.value = result["norm"]
        result.update(self.estimate_npred_excess(datasets=datasets))
        return result



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Pulsar source models."""
import html
import numpy as np
from astropy.units import Quantity

__all__ = ["Pulsar", "SimplePulsar"]

DEFAULT_I = Quantity(1e45, "g cm2")
"""Pulsar default moment of inertia"""

DEFAULT_R = Quantity(1e6, "cm")
"""Pulsar default radius of the neutron star"""

B_CONST = Quantity(3.2e19, "gauss s^(-1/2)")
"""Pulsar default magnetic field constant"""


class SimplePulsar:
    """Magnetic dipole spin-down model for a pulsar.

    Parameters
    ----------
    P : `~astropy.units.Quantity`
        Rotation period (sec).
    P_dot : `~astropy.units.Quantity`
        Rotation period derivative (sec sec^-1).
    I : `~astropy.units.Quantity`
        Moment of inertia (g cm^2).
    R : `~astropy.units.Quantity`
        Radius of the pulsar (cm).
    """

    def __init__(self, P, P_dot, I=DEFAULT_I, R=DEFAULT_R):  # noqa: E741
        self.P = Quantity(P, "s")
        self.P_dot = P_dot
        self.I = I  # noqa: E741
        self.R = R

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    @property
    def luminosity_spindown(self):
        r"""Spin-down luminosity as a `~astropy.units.Quantity`.

        .. math:: \dot{L} = 4\pi^2 I \frac{\dot{P}}{P^{3}}
        """
        return 4 * np.pi**2 * self.I * self.P_dot / self.P**3

    @property
    def tau(self):
        r"""Characteristic age as a `~astropy.units.Quantity`.

        .. math:: \tau = \frac{P}{2\dot{P}}
        """
        return (self.P / (2 * self.P_dot)).to("yr")

    @property
    def magnetic_field(self):
        r"""Magnetic field strength at the polar cap as a `~astropy.units.Quantity`.

        .. math:: B = 3.2 \cdot 10^{19} (P\dot{P})^{1/2} \text{ Gauss}
        """
        return B_CONST * np.sqrt(self.P * self.P_dot)


class Pulsar:
    """Magnetic dipole spin-down pulsar model.

    Parameters
    ----------
    P_0 : float
        Period at birth.
    B : `~astropy.units.Quantity`
        Magnetic field strength at the poles (Gauss).
    n : float
        Spin-down braking index.
    I : float
        Moment of inertia.
    R : float
        Radius.
    """

    def __init__(
        self,
        P_0="0.1 s",
        B="1e10 G",
        n=3,
        I=DEFAULT_I,  # noqa: E741
        R=DEFAULT_R,
        age=None,
        L_0=None,  # noqa: E741
    ):
        P_0 = Quantity(P_0, "s")
        B = Quantity(B, "G")

        self.I = I  # noqa: E741
        self.R = R
        self.P_0 = P_0
        self.B = B
        self.P_dot_0 = (B / B_CONST) ** 2 / P_0
        self.tau_0 = P_0 / (2 * self.P_dot_0)
        self.n = float(n)
        self.beta = -(n + 1.0) / (n - 1.0)
        if age is not None:
            self.age = Quantity(age, "yr")
        if L_0 is None:
            self.L_0 = 4 * np.pi**2 * self.I * self.P_dot_0 / self.P_0**3

    def luminosity_spindown(self, t):
        r"""Spin down luminosity.

        .. math::
            \dot{L}(t) = \dot{L}_0 \left(1 + \frac{t}{\tau_0}\right)^{-\frac{n + 1}{n - 1}}

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the pulsar.
        """
        t = Quantity(t, "yr")
        return self.L_0 * (1 + (t / self.tau_0)) ** self.beta

    def energy_integrated(self, t):
        r"""Total energy released by a given time.

        Time-integrated spin-down luminosity since birth.

        .. math:: E(t) = \dot{L}_0 \tau_0 \frac{t}{t + \tau_0}

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the pulsar.
        """
        t = Quantity(t, "yr")
        return self.L_0 * self.tau_0 * (t / (t + self.tau_0))

    def period(self, t):
        r"""Rotation period.

        .. math::
            P(t) = P_0 \left(1 + \frac{t}{\tau_0}\right)^{\frac{1}{n - 1}}

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the pulsar.
        """
        t = Quantity(t, "yr")
        return self.P_0 * (1 + (t / self.tau_0)) ** (1.0 / (self.n - 1))

    def period_dot(self, t):
        r"""Period derivative at age t.

        P_dot for a given period and magnetic field B, assuming a dipole
        spin-down.

        .. math:: \dot{P}(t) = \frac{B^2}{3.2 \cdot 10^{19} P(t)}

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the pulsar.
        """
        t = Quantity(t, "yr")
        return self.B**2 / (self.period(t) * B_CONST**2)

    def tau(self, t):
        r"""Characteristic age at real age t.

        .. math:: \tau = \frac{P}{2\dot{P}}

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the pulsar.
        """
        t = Quantity(t, "yr")
        return self.period(t) / 2 * self.period_dot(t)

    def magnetic_field(self, t):
        r"""Magnetic field at polar cap (assumed constant).

        .. math::
            B = 3.2 \cdot 10^{19} (P\dot{P})^{1/2} \text{ Gauss}

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the pulsar.
        """
        t = Quantity(t, "yr")
        return B_CONST * np.sqrt(self.period(t) * self.period_dot(t))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Interpolation utilities."""

import html
from itertools import compress
import numpy as np
import scipy.interpolate
from astropy import units as u
from .compat import COPY_IF_NEEDED

__all__ = [
    "interpolate_profile",
    "interpolation_scale",
    "ScaledRegularGridInterpolator",
]

INTERPOLATION_ORDER = {None: 0, "nearest": 0, "linear": 1, "quadratic": 2, "cubic": 3}


class ScaledRegularGridInterpolator:
    """Thin wrapper around `scipy.interpolate.RegularGridInterpolator`.

    The values are scaled before the interpolation and back-scaled after the
    interpolation.

    Dimensions of length 1 are ignored in the interpolation of the data.

    Parameters
    ----------
    points : tuple of `~numpy.ndarray` or `~astropy.units.Quantity`
        Tuple of points passed to `~scipy.interpolate.RegularGridInterpolator`.
    values : `~numpy.ndarray`
        Values passed to `~scipy.interpolate.RegularGridInterpolator`.
    points_scale : tuple of str
        Interpolation scale used for the points.
    values_scale : {'lin', 'log', 'sqrt'}
        Interpolation scaling applied to values. If the values vary over many magnitudes
        a 'log' scaling is recommended.
    axis : int or None
        Axis along which to interpolate.
    method : {"linear", "nearest"}
        Default interpolation method. Can be overwritten when calling the
        `ScaledRegularGridInterpolator`.
    **kwargs : dict
        Keyword arguments passed to `~scipy.interpolate.RegularGridInterpolator`.
    """

    def __init__(
        self,
        points,
        values,
        points_scale=None,
        values_scale="lin",
        extrapolate=True,
        axis=None,
        **kwargs,
    ):
        if points_scale is None:
            points_scale = ["lin"] * len(points)

        self.scale_points = [interpolation_scale(scale) for scale in points_scale]
        self.scale = interpolation_scale(values_scale)
        self.axis = axis

        self._include_dimensions = [len(p) > 1 for p in points]

        values_scaled = self.scale(values)
        points_scaled = self._scale_points(points=points)

        kwargs.setdefault("bounds_error", False)

        if extrapolate:
            kwargs.setdefault("fill_value", None)

        method = kwargs.get("method", None)

        if not np.any(self._include_dimensions):
            if method != "nearest":
                raise ValueError(
                    "Interpolating scalar values requires using "
                    "method='nearest' explicitly."
                )

        if np.any(self._include_dimensions):
            values_scaled = np.squeeze(values_scaled)

        if axis is None:
            self._interpolate = scipy.interpolate.RegularGridInterpolator(
                points=points_scaled, values=values_scaled, **kwargs
            )
        else:
            self._interpolate = scipy.interpolate.interp1d(
                points_scaled[0], values_scaled, axis=axis
            )

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    def _scale_points(self, points):
        points_scaled = [scale(p) for p, scale in zip(points, self.scale_points)]

        if np.any(self._include_dimensions):
            points_scaled = compress(points_scaled, self._include_dimensions)

        return tuple(points_scaled)

    def __call__(self, points, method=None, clip=True, **kwargs):
        """Interpolate data points.

        Parameters
        ----------
        points : tuple of `~numpy.ndarray` or `~astropy.units.Quantity`
            Tuple of coordinate arrays of the form (x_1, x_2, x_3, ...). Arrays are
            broadcast internally.
        method : {None, "linear", "nearest"}
            Linear or nearest neighbour interpolation.
            Default is None, which is `method` defined on init.
        clip : bool
            Clip values at zero after interpolation.
        """
        points = self._scale_points(points=points)

        if self.axis is None:
            points = np.broadcast_arrays(*points)
            points_interp = np.stack([_.flat for _ in points]).T
            values = self._interpolate(points_interp, method, **kwargs)
            values = self.scale.inverse(values.reshape(points[0].shape))
        else:
            values = self._interpolate(points[0])
            values = self.scale.inverse(values)

        if clip:
            values = np.clip(values, 0, np.inf)

        return values


def interpolation_scale(scale="lin"):
    """Interpolation scaling.

    Parameters
    ----------
    scale : {"lin", "log", "sqrt"}
        Choose interpolation scaling.
    """
    if scale in ["lin", "linear"]:
        return LinearScale()
    elif scale == "log":
        return LogScale()
    elif scale == "sqrt":
        return SqrtScale()
    elif scale == "stat-profile":
        return StatProfileScale()
    elif isinstance(scale, InterpolationScale):
        return scale
    else:
        raise ValueError(f"Not a valid value scaling mode: '{scale}'.")


class InterpolationScale:
    """Interpolation scale base class."""

    def __call__(self, values):
        if hasattr(self, "_unit"):
            values = u.Quantity(values, copy=COPY_IF_NEEDED).to_value(self._unit)
        else:
            if isinstance(values, u.Quantity):
                self._unit = values.unit
                values = values.value
        return self._scale(values)

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    def inverse(self, values):
        values = self._inverse(values)
        if hasattr(self, "_unit"):
            return u.Quantity(values, self._unit, copy=COPY_IF_NEEDED)
        else:
            return values

    def _inverse_deriv(self, values):
        values = self._inverse_derivative(values)
        if hasattr(self, "_unit"):
            return u.Quantity(values, self._unit, copy=COPY_IF_NEEDED)
        else:
            return values


class LogScale(InterpolationScale):
    """Logarithmic scaling."""

    tiny = np.finfo(np.float32).tiny

    def _scale(self, values):
        values = np.clip(values, self.tiny, np.inf)
        return np.log(values)

    @classmethod
    def _inverse(cls, values):
        output = np.exp(values)
        return np.where(abs(output) - cls.tiny <= cls.tiny, 0, output)

    @classmethod
    def _inverse_derivative(cls, values):
        return cls._inverse(values)


class SqrtScale(InterpolationScale):
    """Square root scaling."""

    @staticmethod
    def _scale(values):
        sign = np.sign(values)
        return sign * np.sqrt(sign * values)

    @classmethod
    def _inverse(cls, values):
        return np.power(values, 2)

    @classmethod
    def _inverse_derivative(cls, values):
        return 2.0 * values


class StatProfileScale(InterpolationScale):
    """Square root profile scaling."""

    def __init__(self, axis=0):
        self.axis = axis

    def _scale(self, values):
        values = np.sign(np.gradient(values, axis=self.axis)) * values
        sign = np.sign(values)
        return sign * np.sqrt(sign * values)

    @classmethod
    def _inverse(cls, values):
        return np.power(values, 2)


class LinearScale(InterpolationScale):
    """Linear scaling."""

    @staticmethod
    def _scale(values):
        return values

    @classmethod
    def _inverse(cls, values):
        return values

    @classmethod
    def _inverse_derivative(cls, values):
        return 1.0


def interpolate_profile(x, y, interp_scale="sqrt", extrapolate=False):
    """Helper function to interpolate one-dimensional profiles.

    Parameters
    ----------
    x : `~numpy.ndarray`
        Array of x values.
    y : `~numpy.ndarray`
        Array of y values.
    interp_scale : {"sqrt", "lin"}
        Interpolation scale applied to the profile. If the profile is
        of parabolic shape, a "sqrt" scaling is recommended. In other cases or
        for fine sampled profiles a "lin" can also be used.
        Default is "sqrt".
    extrapolate : bool
        Extrapolate or not if the evaluation value is outside the range of x values.
        Default is False.

    Returns
    -------
    interp : `interp1d`
        Interpolator.
    """
    method_dict = {"sqrt": "quadratic", "lin": "linear"}
    kwargs = dict(kind=method_dict[interp_scale])
    if extrapolate:
        kwargs["bounds_error"] = False
        kwargs["fill_value"] = "extrapolate"
    return scipy.interpolate.interp1d(x, y, **kwargs)
